import torch
import triton
import triton.language as tl
import time
import os
os.environ['ENABLE_TMA'] = '1'


@triton.jit
def grouped_launch(pid,
                m, n,
                block_m: tl.constexpr, block_n: tl.constexpr, group_m: tl.constexpr):
    
    grid_m = tl.cdiv(m, block_m)
    grid_n = tl.cdiv(n, block_n)

    width = group_m * grid_n
    group_id = pid // width
    group_size = tl.minimum(grid_m - group_id * group_m, group_m)

    pid_m = group_id * group_m + (pid % group_size)
    pid_n = (pid % width) // group_size

    return pid_m, pid_n

@triton.jit()
def column_major(pid,
              m, n,
              block_m: tl.constexpr, block_n: tl.constexpr):
    
    grid_m = tl.cdiv(m, block_m) 

    pid_m = pid % grid_m
    pid_n = pid // grid_m

    return pid_m, pid_n

@triton.jit
def scaled_gemm_splitk(a_ptr, b_ptr, c_ptr,
            stride_am, stride_ak,
            stride_bk, stride_bn,
            stride_cm, stride_cn,
            scale_a, scale_b,
            m, n, k,
            block_m: tl.constexpr, block_n: tl.constexpr, block_k: tl.constexpr,
            split_k: tl.constexpr, group_m: tl.constexpr):
    
    pid = tl.program_id(0)
    pid_k = tl.program_id(1)
    grid_k = tl.cdiv(k, block_k*split_k)

    # Column Major produces speedup over Grouped Launch for small-to-medium M
    pid_m, pid_n = column_major(pid,
                                m, n,
                                block_m, block_n)


    offs_m = pid_m*block_m + tl.arange(0, block_m)
    offs_n = pid_n*block_n + tl.arange(0, block_n)
    offs_k = pid_k*block_k + tl.arange(0, block_k)

    offs_am = tl.max_contiguous(tl.multiple_of(offs_m, block_m), block_m)
    offs_bn = tl.max_contiguous(tl.multiple_of(offs_n, block_n), block_n)

    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)


    acc = tl.zeros((block_m, block_n), dtype=tl.float32)
    for k_ in range(0, grid_k):
        
        k_remaining = k - k_ * (block_k * split_k)

        a = tl.load(a_ptrs, mask=offs_k[None, :] < k_remaining, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < k_remaining, other=0.0)

        acc = tl.dot(a, b, acc, out_dtype=tl.float32)

        a_ptrs += block_k * split_k * stride_ak
        b_ptrs += block_k * split_k * stride_bk
    
    # Scaled in SRAM before write back to DRAM
    acc = scale_a * scale_b * acc
    acc.to(tl.float16)

    offs_m = pid_m*block_m + tl.arange(0, block_m)
    offs_n = pid_n*block_n + tl.arange(0, block_n)
    
    c_ptrs = c_ptr + (offs_m[:, None] * stride_cm + offs_n[None, :] * stride_cn)
    mask = (offs_m < m)[:, None] & (offs_n < n)[None, :]
    
    tl.atomic_add(c_ptrs, acc, mask=mask)

def scaled_mm_splitk(a, b, scale_a: float=1.0, scale_b: float=1.0):
    assert a.shape[1] == b.shape[0]
    m, k = a.shape
    _, n = b.shape
    
    block_m = 64
    block_n = 64
    block_k = 256
    num_stages = 3
    num_warps = 8
    split_k = 4
    group_m = 8

    total_blocks_m = triton.cdiv(m, block_m)
    total_blocks_n = triton.cdiv(n, block_n)
    total_programs_mn = total_blocks_m * total_blocks_n
    total_programs_k = split_k
    
    grid = (total_programs_mn, total_programs_k)

    c = torch.zeros((m, n), device=a.device, dtype=torch.float16)
    k = scaled_gemm_splitk[grid](a, b, c,
                              a.stride(0), a.stride(1),
                              b.stride(0), b.stride(1),
                              c.stride(0), c.stride(1),
                              scale_a, scale_b,                              
                              m, n, k,
                              block_m, block_n, block_k,
                              split_k, group_m, num_stages=num_stages, num_warps=num_warps)

    return c


# Copyright (c) 2023-2024, Songlin Yang, Yu Zhang

import torch
import triton
import triton.language as tl

from fla.ops.utils.op import exp
from fla.utils import IS_AMD, autotune_cache_kwargs

NUM_WARPS_AUTOTUNE = [1, 2, 4, 8, 16] if IS_AMD else [1, 2, 4, 8, 16, 32]


@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in NUM_WARPS_AUTOTUNE
    ],
    key=['D'],
    **autotune_cache_kwargs,
)
@triton.jit
def softmax_fwd_kernel(
    x,
    p,
    D: tl.constexpr,
    B: tl.constexpr,
):
    i_n = tl.program_id(0)
    o_d = tl.arange(0, B)
    m_d = o_d < D

    b_x = tl.load(x + i_n * D + o_d, mask=m_d, other=-float('inf'))
    b_m = tl.max(b_x, 0)
    b_x = exp(b_x - b_m)
    b_p = b_x / tl.sum(b_x, 0)

    tl.store(p + i_n * D + o_d, b_p.to(p.dtype.element_ty), mask=m_d)


@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in NUM_WARPS_AUTOTUNE
    ],
    key=['D'],
    **autotune_cache_kwargs,
)
@triton.jit
def softmax_bwd_kernel(
    p,
    dp,
    ds,
    D: tl.constexpr,
    B: tl.constexpr,
):
    i_n = tl.program_id(0)
    o_d = tl.arange(0, B)
    m_d = o_d < D

    b_p = tl.load(p + i_n * D + o_d, mask=m_d, other=0.)
    b_dp = tl.load(dp + i_n * D + o_d, mask=m_d, other=0.)
    b_pp = tl.sum(b_p * b_dp, 0)
    b_ds = b_p * b_dp - b_p * b_pp
    tl.store(ds + i_n * D + o_d, b_ds.to(ds.dtype.element_ty), mask=m_d)


def softmax_fwd(
    x: torch.Tensor,
    dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    shape = x.shape
    x = x.view(-1, x.shape[-1])

    N, D = x.shape
    B = triton.next_power_of_2(D)

    p = torch.empty_like(x, dtype=dtype)
    softmax_fwd_kernel[(N,)](
        x=x,
        p=p,
        D=D,
        B=B,
    )
    return p.view(*shape)


def softmax_bwd(
    p: torch.Tensor,
    dp: torch.Tensor,
    dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    shape = p.shape
    p = p.view(-1, p.shape[-1])
    ds = torch.empty_like(p, dtype=dtype)

    N, D = p.shape
    B = triton.next_power_of_2(D)
    softmax_bwd_kernel[(N,)](
        p=p,
        dp=dp,
        ds=ds,
        D=D,
        B=B,
    )
    return ds.view(*shape)



import logging
import math
from collections import namedtuple

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle
from flag_gems.utils.limits import get_dtype_max

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@libentry()
@triton.jit
def min_kernel_1(
    inp,
    mid,
    M,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    max_value = get_dtype_max(inp.type.element_ty)
    inp_val = tl.load(inp_ptrs, mask=mask, other=max_value)
    min_val = tl.min(inp_val)
    mid_ptr = mid + pid
    tl.store(mid_ptr, min_val)


@libentry()
@triton.jit
def min_kernel_2(mid, out, mid_size, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid + offset
    mask = offset < mid_size
    max_value = get_dtype_max(mid.type.element_ty)
    mid_val = tl.load(mid_ptrs, mask=mask, other=max_value)
    min_val = tl.min(mid_val)
    tl.store(out, min_val)


def heur_block_n(args):
    return triton.next_power_of_2(args["N"])


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("min"),
    key=[
        "M",
        "N",
    ],
)
@triton.jit
def min_kernel(
    inp,
    out_value,
    out_index,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # set offset
    pid_m = tle.program_id(0)
    pid_k = tle.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

    dtype = inp.type.element_ty
    # you just cannot create a function that return a tl.dtype in triton lang
    acc_type = tl.float32 if dtype is tl.bfloat16 else dtype
    max_value = get_dtype_max(dtype)
    min_values = tl.full([BLOCK_M], dtype=acc_type, value=max_value)
    argmin_values = tl.full([BLOCK_M], dtype=tl.int64, value=0)
    for start_n in range(0, N, BLOCK_N):
        n_offset = start_n + tl.arange(0, BLOCK_N)
        offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
        mask = m_offset[:, None] < M and n_offset[None, :] < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=max_value)
        if dtype is tl.int64:
            inp_vals = tl.where(mask, inp_vals, max_value)
        local_min, local_argmin = tl.min(inp_vals, 1, return_indices=True)
        # if return indices is not supported, call a tl.argmax in addition
        # local_argmin = tl.argmin(inp_vals, 1)
        update = local_min < min_values
        min_values = tl.where(update, local_min, min_values)
        argmin_values = tl.where(update, start_n + local_argmin, argmin_values)

    offset_index = m_offset * K + pid_k
    out_value_ptrs = out_value + offset_index
    out_index_ptrs = out_index + offset_index
    mask1 = m_offset < M
    tl.store(out_value_ptrs, min_values, mask=mask1)
    tl.store(out_index_ptrs, argmin_values, mask=mask1)


def min(inp):
    logger.debug("GEMS_ASCEND MIN")
    M = inp.numel()
    block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
    mid_size = triton.cdiv(M, block_size)
    block_mid = triton.next_power_of_2(mid_size)

    dtype = inp.dtype
    mid = torch.empty((mid_size,), dtype=dtype, device=inp.device)
    out = torch.empty([], dtype=dtype, device=inp.device)

    with torch_device_fn.device(inp.device):
        min_kernel_1[(mid_size, 1, 1)](inp, mid, M, block_size)
        min_kernel_2[(1, 1, 1)](mid, out, mid_size, block_mid)
    return out


def min_dim(inp, dim=None, keepdim=False):
    logger.debug("GEMS_ASCEND MIN DIM")
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    shape = inp.shape
    dim = dim % inp.ndim
    N = shape[dim]
    M = math.prod(shape[:dim])
    K = inp.numel() // M // N

    inp = inp.contiguous()

    shape_list = list(shape)
    shape_list[dim] = 1
    out_value = torch.empty(shape_list, dtype=inp.dtype, device=inp.device)
    out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)

    if not keepdim:
        out_value = torch.squeeze(out_value, dim)
        out_index = torch.squeeze(out_index, dim)

    grid = lambda meta: (
        triton.cdiv(M, meta["BLOCK_M"]),
        K,
    )
    with torch_device_fn.device(inp.device):
        min_kernel[grid](inp, out_value, out_index, M, N, K)
    Min_out = namedtuple("min", ["values", "indices"])
    out = Min_out(values=out_value, indices=out_index)
    return out



import logging

import triton
import triton.language as tl
from _kunlunxin.utils.codegen_config_utils import CodeGenConfig

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger(__name__)

config_ = CodeGenConfig(
    512,
    (65536, 65536, 65536),
    32,
    True,
    prefer_1d_tile=True,
    buffer_size_limit=4096,
    isCloseVectorization=True,
    unroll_num=8,
)


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")], config=config_)
@triton.jit
def reciprocal_func(x):
    return 1.0 / x.to(tl.float32)


def reciprocal(A):
    logger.debug("GEMS RECIPROCAL")
    return reciprocal_func(A)


def reciprocal_(A):
    logger.debug("GEMS RECIPROCAL_")
    return reciprocal_func(A, out0=A)



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch
import triton
import triton.language as tl

from fla.ops.utils.index import prepare_chunk_indices
from fla.utils import autotune_cache_kwargs, check_shared_mem, input_guard

BS_LIST = [32, 64] if check_shared_mem() else [16, 32]


@triton.heuristics({
    'HAS_SCALE': lambda args: args['scale'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4, 8]
    ],
    key=['B', 'H', 'BT', 'IS_VARLEN', 'REVERSE'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_local_cumsum_scalar_kernel(
    s,
    o,
    scale,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    BT: tl.constexpr,
    REVERSE: tl.constexpr,
    HAS_SCALE: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    if HEAD_FIRST:
        p_s = tl.make_block_ptr(s + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
        p_o = tl.make_block_ptr(o + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
    else:
        p_s = tl.make_block_ptr(s + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
        p_o = tl.make_block_ptr(o + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
    # [BT]
    b_s = tl.load(p_s, boundary_check=(0,)).to(tl.float32)
    b_o = tl.cumsum(b_s, axis=0)
    if REVERSE:
        b_z = tl.sum(b_s, axis=0)
        b_o = -b_o + b_z[None] + b_s
    if HAS_SCALE:
        b_o *= scale
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0,))


@triton.heuristics({
    'HAS_SCALE': lambda args: args['scale'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BS': BS}, num_warps=num_warps)
        for BS in BS_LIST
        for num_warps in [2, 4, 8]
    ],
    key=['B', 'H', 'S', 'BT', 'IS_VARLEN', 'REVERSE'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_local_cumsum_vector_kernel(
    s,
    o,
    scale,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    S: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    REVERSE: tl.constexpr,
    HAS_SCALE: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_s, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    if HEAD_FIRST:
        p_s = tl.make_block_ptr(s + (bos * H + i_h*T)*S, (T, S), (S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
        p_o = tl.make_block_ptr(o + (bos * H + i_h*T)*S, (T, S), (S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
    else:
        p_s = tl.make_block_ptr(s + (bos * H + i_h) * S, (T, S), (H*S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
        p_o = tl.make_block_ptr(o + (bos * H + i_h) * S, (T, S), (H*S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
    # [BT, BS]
    b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
    if REVERSE:
        b_o = tl.cumsum(b_s, axis=0, reverse=True)
    else:
        b_o = tl.cumsum(b_s, axis=0)
    if HAS_SCALE:
        b_o *= scale
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))



@triton.heuristics({
    'HAS_SCALE': lambda args: args['scale'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BT': BT}, num_warps=num_warps, num_stages=num_stages)
        for BT in [32, 64, 128, 256]
        for num_warps in [2, 4, 8]
        for num_stages in [1, 2, 3, 4]
    ],
    key=['B', 'H', 'IS_VARLEN', 'REVERSE'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_global_cumsum_scalar_kernel(
    s,
    o,
    scale,
    cu_seqlens,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    BT: tl.constexpr,
    REVERSE: tl.constexpr,
    HAS_SCALE: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_nh = tl.program_id(0)
    i_n, i_h = i_nh // H, i_nh % H
    if IS_VARLEN:
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
    else:
        bos, eos = i_n * T, i_n * T + T
    T = eos - bos

    b_z = tl.zeros([], dtype=tl.float32)
    NT = tl.cdiv(T, BT)
    for i_c in range(NT):
        i_t = NT - 1 - i_c if REVERSE else i_c
        if HEAD_FIRST:
            p_s = tl.make_block_ptr(s + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
            p_o = tl.make_block_ptr(o + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
        else:
            p_s = tl.make_block_ptr(s + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
            p_o = tl.make_block_ptr(o + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
        b_s = tl.load(p_s, boundary_check=(0,)).to(tl.float32)
        b_o = tl.cumsum(b_s, axis=0)
        b_ss = tl.sum(b_s, 0)
        if REVERSE:
            b_o = -b_o + b_ss + b_s
        b_o += b_z
        if i_c >= 0:
            b_z += b_ss
        if HAS_SCALE:
            b_o *= scale
        tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0,))


@triton.heuristics({
    'HAS_SCALE': lambda args: args['scale'] is not None,
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BT': BT}, num_warps=num_warps, num_stages=num_stages)
        for BT in [16, 32, 64, 128]
        for num_warps in [2, 4, 8]
        for num_stages in [1, 2, 3, 4]
    ],
    key=['B', 'H', 'S', 'IS_VARLEN', 'REVERSE'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_global_cumsum_vector_kernel(
    s,
    o,
    scale,
    cu_seqlens,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    S: tl.constexpr,
    BT: tl.constexpr,
    BS: tl.constexpr,
    REVERSE: tl.constexpr,
    HAS_SCALE: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_s, i_nh = tl.program_id(0), tl.program_id(1)
    i_n, i_h = i_nh // H, i_nh % H
    if IS_VARLEN:
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
    else:
        bos, eos = i_n * T, i_n * T + T
    T = eos - bos

    b_z = tl.zeros([BS], dtype=tl.float32)
    NT = tl.cdiv(T, BT)
    for i_c in range(NT):
        i_t = NT - 1 - i_c if REVERSE else i_c
        if HEAD_FIRST:
            p_s = tl.make_block_ptr(s + (bos * H + i_h*T)*S, (T, S), (S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
            p_o = tl.make_block_ptr(o + (bos * H + i_h*T)*S, (T, S), (S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
        else:
            p_s = tl.make_block_ptr(s + (bos * H + i_h) * S, (T, S), (H*S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
            p_o = tl.make_block_ptr(o + (bos * H + i_h) * S, (T, S), (H*S, 1), (i_t * BT, i_s * BS), (BT, BS), (1, 0))
        # [BT, BS]
        b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
        if REVERSE:
            b_c = b_z[None, :] + tl.cumsum(b_s, axis=0, reverse=True)
        else:
            b_c = b_z[None, :] + tl.cumsum(b_s, axis=0)
        if HAS_SCALE:
            b_c *= scale
        tl.store(p_o, b_c.to(p_o.dtype.element_ty), boundary_check=(0, 1))
        b_z += tl.sum(b_s, 0)


def chunk_local_cumsum_scalar(
    g: torch.Tensor,
    chunk_size: int,
    reverse: bool = False,
    scale: float = None,
    cu_seqlens: torch.Tensor | None = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
    chunk_indices: torch.LongTensor | None = None,
) -> torch.Tensor:
    if head_first:
        B, H, T = g.shape
    else:
        B, T, H = g.shape
    assert chunk_size == 2**(chunk_size.bit_length()-1), "chunk_size must be a power of 2"
    BT = chunk_size
    if chunk_indices is None and cu_seqlens is not None:
        chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    g_org, g = g, torch.empty_like(g, dtype=output_dtype or g.dtype)
    grid = (NT, B * H)
    chunk_local_cumsum_scalar_kernel[grid](
        s=g_org,
        o=g,
        scale=scale,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        T=T,
        B=B,
        H=H,
        BT=BT,
        HEAD_FIRST=head_first,
        REVERSE=reverse,
    )
    return g


def chunk_local_cumsum_vector(
    g: torch.Tensor,
    chunk_size: int,
    reverse: bool = False,
    scale: float = None,
    cu_seqlens: torch.Tensor | None = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
    chunk_indices: torch.LongTensor | None = None,
) -> torch.Tensor:
    if head_first:
        B, H, T, S = g.shape
    else:
        B, T, H, S = g.shape
    BT = chunk_size
    if chunk_indices is None and cu_seqlens is not None:
        chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    assert chunk_size == 2**(chunk_size.bit_length()-1), "chunk_size must be a power of 2"

    g_org, g = g, torch.empty_like(g, dtype=output_dtype or g.dtype)
    def grid(meta): return (triton.cdiv(meta['S'], meta['BS']), NT, B * H)
    # keep cummulative normalizer in fp32
    # this kernel is equivalent to
    # g = g.view(B, H, NT, BT, -1).cumsum(-2).view(B, H, T, -1)
    chunk_local_cumsum_vector_kernel[grid](
        s=g_org,
        o=g,
        scale=scale,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        T=T,
        B=B,
        H=H,
        S=S,
        BT=BT,
        HEAD_FIRST=head_first,
        REVERSE=reverse,
    )
    return g


@input_guard
def chunk_global_cumsum_scalar(
    s: torch.Tensor,
    reverse: bool = False,
    cu_seqlens: torch.Tensor | None = None,
    scale: float = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    if head_first:
        B, H, T = s.shape
    else:
        B, T, H = s.shape
    N = len(cu_seqlens) - 1 if cu_seqlens is not None else B

    z = torch.empty_like(s, dtype=output_dtype or s.dtype)
    grid = (N * H,)
    chunk_global_cumsum_scalar_kernel[grid](
        s=s,
        o=z,
        scale=scale,
        cu_seqlens=cu_seqlens,
        T=T,
        B=B,
        H=H,
        HEAD_FIRST=head_first,
        REVERSE=reverse,
    )
    return z


@input_guard
def chunk_global_cumsum_vector(
    s: torch.Tensor,
    reverse: bool = False,
    cu_seqlens: torch.Tensor | None = None,
    scale: float = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    if head_first:
        B, H, T, S = s.shape
    else:
        B, T, H, S = s.shape
    N = len(cu_seqlens) - 1 if cu_seqlens is not None else B
    BS = min(32, triton.next_power_of_2(S))

    z = torch.empty_like(s, dtype=output_dtype or s.dtype)
    grid = (triton.cdiv(S, BS), N * H)
    chunk_global_cumsum_vector_kernel[grid](
        s=s,
        o=z,
        scale=scale,
        cu_seqlens=cu_seqlens,
        T=T,
        B=B,
        H=H,
        S=S,
        BS=BS,
        HEAD_FIRST=head_first,
        REVERSE=reverse,
    )
    return z


@input_guard
def chunk_global_cumsum(
    s: torch.Tensor,
    reverse: bool = False,
    cu_seqlens: torch.Tensor | None = None,
    scale: float = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    if cu_seqlens is not None:
        assert s.shape[0] == 1, "Only batch size 1 is supported when cu_seqlens are provided"
    if len(s.shape) == 3:
        return chunk_global_cumsum_scalar(
            s=s,
            reverse=reverse,
            cu_seqlens=cu_seqlens,
            scale=scale,
            head_first=head_first,
            output_dtype=output_dtype,
        )
    elif len(s.shape) == 4:
        return chunk_global_cumsum_vector(
            s=s,
            reverse=reverse,
            cu_seqlens=cu_seqlens,
            scale=scale,
            head_first=head_first,
            output_dtype=output_dtype,
        )
    else:
        raise ValueError(
            f"Unsupported input shape {s.shape}, "
            f"which should be [B, T, H]/[B, T, H, D] if `head_first=False` "
            f"or [B, H, T]/[B, H, T, D] otherwise",
        )


@input_guard
def chunk_local_cumsum(
    g: torch.Tensor,
    chunk_size: int,
    reverse: bool = False,
    scale: float = None,
    cu_seqlens: torch.Tensor | None = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
    chunk_indices: torch.LongTensor | None = None,
    **kwargs,
) -> torch.Tensor:
    if cu_seqlens is not None:
        assert g.shape[0] == 1, "Only batch size 1 is supported when cu_seqlens are provided"
    if len(g.shape) == 3:
        return chunk_local_cumsum_scalar(
            g=g,
            chunk_size=chunk_size,
            reverse=reverse,
            scale=scale,
            cu_seqlens=cu_seqlens,
            head_first=head_first,
            output_dtype=output_dtype,
            chunk_indices=chunk_indices,
        )
    elif len(g.shape) == 4:
        return chunk_local_cumsum_vector(
            g=g,
            chunk_size=chunk_size,
            reverse=reverse,
            scale=scale,
            cu_seqlens=cu_seqlens,
            head_first=head_first,
            output_dtype=output_dtype,
            chunk_indices=chunk_indices,
        )
    else:
        raise ValueError(
            f"Unsupported input shape {g.shape}, "
            f"which should be (B, T, H, D) if `head_first=False` "
            f"or (B, H, T, D) otherwise",
        )



import logging
import math

import torch
import triton

import flag_gems
from flag_gems import runtime
from flag_gems.ops.flash_kernel import (
    block_m_splitkv_heuristic,
    block_n_splitkv_heuristic,
    flash_fwd_kernel,
    flash_fwd_splitkv_combine_kernel,
    flash_fwd_splitkv_kernel,
    flash_varlen_fwd_kernel,
)
from flag_gems.runtime import torch_device_fn
from flag_gems.utils.random_utils import philox_backend_seed_offset

logger = logging.getLogger(__name__)
_debug = False


def CHECK_DEVICE(x):
    assert x.device.type == flag_gems.device


class fwd_params:
    __slots__ = (
        # pointers and strides
        "q_ptr",
        "k_ptr",
        "v_ptr",
        "o_ptr",
        "p_ptr",
        "softmax_lse_ptr",
        "q_row_stride",
        "k_row_stride",
        "v_row_stride",
        "q_head_stride",
        "k_head_stride",
        "v_head_stride",
        "o_row_stride",
        "o_head_stride",
        "q_batch_stride",
        "k_batch_stride",
        "v_batch_stride",
        "o_batch_stride",
        "is_cu_seqlens_q",
        "cu_seqlens_q_ptr",
        "is_cu_seqlens_k",
        "cu_seqlens_k_ptr",
        "is_seqused_k",
        "seqused_k_ptr",
        # sizes
        "b",
        "bk",
        "h",
        "hk",
        "h_hk_ratio",
        "seqlen_q",
        "seqlen_k",
        "seqlen_q_rounded",
        "seqlen_k_rounded",
        "d",
        "d_rounded",
        # scaling factors
        "is_softcap",
        "softcap",
        "scale_softmax",
        "scale_softmax_log2",
        # dropout
        "is_dropout",
        "p_dropout",
        "rp_dropout",
        "p_dropout_in_uint8_t",
        "philox_args",
        "return_softmax",
        # masking
        "is_causal",
        "is_local",
        "window_size_left",
        "window_size_right",
        "seqlenq_ngroups_swapped",
        # alibi
        "is_alibi",
        "alibi_slopes_ptr",
        "alibi_slopes_batch_stride",
        # block table
        "total_q",
        "page_table_ptr",
        "page_table_batch_stride",
        "block_size",
    )

    def __init__(
        self,
        q_ptr,
        k_ptr,
        v_ptr,
        o_ptr,
        p_ptr,
        softmax_lse_ptr,
        q_row_stride,
        k_row_stride,
        v_row_stride,
        q_head_stride,
        k_head_stride,
        v_head_stride,
        o_row_stride,
        o_head_stride,
        q_batch_stride,
        k_batch_stride,
        v_batch_stride,
        o_batch_stride,
        is_cu_seqlens_q,
        cu_seqlens_q_ptr,
        is_cu_seqlens_k,
        cu_seqlens_k_ptr,
        is_seqused_k,
        seqused_k_ptr,
        # sizes
        b,
        bk,
        h,
        hk,
        h_hk_ratio,
        seqlen_q,
        seqlen_k,
        seqlen_q_rounded,
        seqlen_k_rounded,
        d,
        d_rounded,
        # scaling factors
        is_softcap,
        softcap,
        scale_softmax,
        scale_softmax_log2,
        # dropout
        is_dropout,
        p_dropout,
        rp_dropout,
        p_dropout_in_uint8_t,
        philox_args,
        return_softmax,
        # masking
        is_causal,
        is_local,
        window_size_left,
        window_size_right,
        seqlenq_ngroups_swapped,
        # alibi
        is_alibi,
        alibi_slopes_ptr,
        alibi_slopes_batch_stride,
        # block table
        total_q,
        page_table_ptr,
        page_table_batch_stride,
        block_size,
    ):
        self.q_ptr = q_ptr
        self.k_ptr = k_ptr
        self.v_ptr = v_ptr
        self.o_ptr = o_ptr
        self.p_ptr = p_ptr
        self.softmax_lse_ptr = softmax_lse_ptr
        self.q_row_stride = q_row_stride
        self.k_row_stride = k_row_stride
        self.v_row_stride = v_row_stride
        self.q_head_stride = q_head_stride
        self.k_head_stride = k_head_stride
        self.v_head_stride = v_head_stride
        self.o_row_stride = o_row_stride
        self.o_head_stride = o_head_stride
        self.q_batch_stride = q_batch_stride
        self.k_batch_stride = k_batch_stride
        self.v_batch_stride = v_batch_stride
        self.o_batch_stride = o_batch_stride
        self.is_cu_seqlens_q = is_cu_seqlens_q
        self.cu_seqlens_q_ptr = cu_seqlens_q_ptr
        self.is_cu_seqlens_k = is_cu_seqlens_k
        self.cu_seqlens_k_ptr = cu_seqlens_k_ptr
        self.is_seqused_k = is_seqused_k
        self.seqused_k_ptr = seqused_k_ptr
        # sizes
        self.b = b
        self.bk = bk
        self.h = h
        self.hk = hk
        self.h_hk_ratio = h_hk_ratio
        self.seqlen_q = seqlen_q
        self.seqlen_k = seqlen_k
        self.seqlen_q_rounded = seqlen_q_rounded
        self.seqlen_k_rounded = seqlen_k_rounded
        self.d = d
        self.d_rounded = d_rounded
        # scaling factors
        self.is_softcap = is_softcap
        self.softcap = softcap
        self.scale_softmax = scale_softmax
        self.scale_softmax_log2 = scale_softmax_log2
        # dropout
        self.is_dropout = is_dropout
        self.p_dropout = p_dropout
        self.rp_dropout = rp_dropout
        self.p_dropout_in_uint8_t = p_dropout_in_uint8_t
        self.philox_args = philox_args
        self.return_softmax = return_softmax
        # masking
        self.is_causal = is_causal
        self.is_local = is_local
        self.window_size_left = window_size_left
        self.window_size_right = window_size_right
        self.seqlenq_ngroups_swapped = seqlenq_ngroups_swapped
        # alibi
        self.is_alibi = is_alibi
        self.alibi_slopes_ptr = alibi_slopes_ptr
        self.alibi_slopes_batch_stride = alibi_slopes_batch_stride
        # block table
        self.total_q = total_q
        self.page_table_ptr = page_table_ptr
        self.page_table_batch_stride = page_table_batch_stride
        self.block_size = block_size

    def args(self):
        return tuple(getattr(self, k) for k in self.__slots__)


def mha_varlan_fwd(
    q,
    k,
    v,
    out,
    cu_seqlens_q,
    cu_seqlens_k,
    seqused_k,
    leftpad_k,
    page_table,
    alibi_slopes,
    max_seqlen_q,
    max_seqlen_k,
    p_dropout,
    softmax_scale,
    zero_tensors,
    is_causal,
    window_size_left,
    window_size_right,
    softcap,
    return_softmax,
    gen,
):
    CHECK_DEVICE(q), CHECK_DEVICE(k), CHECK_DEVICE(v)
    q_device = q.device
    q_dtype = q.dtype
    assert q_dtype in (
        torch.float16,
        torch.bfloat16,
    ), "FlashAttention only support fp16 and bf16 data type"
    assert q_dtype == k.dtype
    assert q_dtype == v.dtype
    assert q.stride(-1) == 1, "Input tensor must have contiguous last dimension"
    assert k.stride(-1) == 1, "Input tensor must have contiguous last dimension"
    assert v.stride(-1) == 1, "Input tensor must have contiguous last dimension"

    assert cu_seqlens_q.dtype == torch.int32
    assert cu_seqlens_q.is_contiguous()

    assert cu_seqlens_k.dtype == torch.int32
    assert cu_seqlens_k.is_contiguous()

    assert page_table is not None

    # q shape: [total_q_tokens, num_heads, head_size]
    # k shape:
    #   paged_kv: [num_pages, block_size, num_heads_k, head_size]
    # batch_size, number of sentences
    total_q, num_heads, head_size = q.size()
    num_heads_k = k.size(2)
    batch_size = cu_seqlens_q.numel() - 1
    block_size = k.size(1)
    num_pages = k.size(0)
    k_batch_size = num_pages
    # max_num_pages_per_seq = page_table.size(1)
    page_table_batch_stride = page_table.stride(0)
    k_batch_stride = k.stride(0)
    v_batch_stride = v.stride(0)

    assert k.size() == v.size()
    assert cu_seqlens_q.size() == (batch_size + 1,)
    assert cu_seqlens_k.size() == (batch_size + 1,)

    # Check output shape
    if out is not None:
        assert out.stride(-1) == 1
        assert out.dtype == q.dtype
        assert out.size() == (total_q, num_heads, head_size)

    if seqused_k is not None:
        assert seqused_k.is_contiguous()
        assert seqused_k.size() == (batch_size,)

    if max_seqlen_q == 1 and alibi_slopes is None:
        is_causal = False

    if is_causal:
        window_size_right = 0

    # check disable swa
    if window_size_left >= max_seqlen_k:
        window_size_left = -1
    if window_size_right >= max_seqlen_k:
        window_size_right = -1

    is_local = window_size_left >= 0

    # Optimize all single-query sequences by swapping the query-group and sequence dimensions
    seqlenq_ngroups_swapped = (
        max_seqlen_q == 1
        and alibi_slopes is None
        and num_heads > num_heads_k
        and window_size_left < 0
        and window_size_right < 0
        and p_dropout == 0
    )
    q_groups = num_heads // num_heads_k
    if seqlenq_ngroups_swapped:
        logger.debug("Swapping query groups and sequence dimensions")
        q = (
            q.reshape((batch_size, num_heads_k, q_groups, head_size))
            .transpose(1, 2)
            .reshape(batch_size * q_groups, num_heads_k, head_size)
        )
        max_seqlen_q = q_groups
        num_heads = num_heads_k
        cu_seqlens_q = None
        q_batch_stride = q.stride(0) * max_seqlen_q
        k_batch_stride = k.stride(0)
        v_batch_stride = v.stride(0)
        # o_batch_stride = out.stride(0) * max_seqlen_q
    else:
        q_batch_stride = 0
        k_batch_stride = 0
        v_batch_stride = 0
        o_batch_stride = 0

    total_q = q.size(0)

    assert leftpad_k is None, "leftpad_k is not supported."
    assert (
        head_size <= 256
    ), "FlashAttention forward only supports head dimension at most 256"
    assert (
        head_size % 8 == 0
    ), "head_size must be a multiple of 8, this is ensured by padding!"
    assert (
        num_heads % num_heads_k == 0
    ), "Number of heads in key/value must divide number of heads in query"

    assert q.shape == (total_q, num_heads, head_size)
    assert k.shape == (num_pages, block_size, num_heads_k, head_size)
    assert v.shape == (num_pages, block_size, num_heads_k, head_size)
    assert k.stride() == v.stride()

    if softcap > 0.0:
        assert p_dropout == 0, "dropout is not supported if softcap is used."

    round_multiple = lambda x, m: (x + m - 1) // m * m
    head_size_rounded = round_multiple(head_size, 32) if head_size <= 192 else 256
    seqlen_q_rounded = round_multiple(max_seqlen_q, 128)
    seqlen_k_rounded = round_multiple(max_seqlen_k, 32)

    M_LOG2E = 1.4426950408889634074
    if softcap > 0.0:
        is_softcap = True
        adjusted_scale_softmax = softcap
        adjusted_softcap = softmax_scale / softcap
        adjusted_scale_softmax_log2e = softcap * M_LOG2E
    else:
        is_softcap = False
        adjusted_softcap = 0.0
        adjusted_scale_softmax = softmax_scale
        adjusted_scale_softmax_log2e = softmax_scale * M_LOG2E

    # Set alibi params
    if alibi_slopes is not None:
        assert alibi_slopes.device == q_device
        assert alibi_slopes.dtype in (torch.float,)
        assert alibi_slopes.stride(-1) == 1
        assert alibi_slopes.shape == (num_heads,) or alibi_slopes.shape == (
            batch_size,
            num_heads,
        )
        alibi_slopes_batch_stride = (
            alibi_slopes.stride(0) if alibi_slopes.ndim == 2 else 0
        )
        is_alibi = True
    else:
        alibi_slopes_batch_stride = 0
        is_alibi = False

    # Prepare params to kernel
    with torch_device_fn.device(q_device):
        if out is not None:
            out_ = out
            if seqlenq_ngroups_swapped:
                out = torch.empty_like(q, dtype=v.dtype)
        else:
            out_ = None
            out = torch.empty_like(q, dtype=v.dtype)

        if seqlenq_ngroups_swapped:
            o_batch_stride = out.stride(0) * max_seqlen_q

        lse = torch.empty((num_heads, total_q), dtype=torch.float, device=q_device)

        if p_dropout > 0:
            is_dropout = True
            increment = batch_size * num_heads * 32
            philox_seed, philox_offset = philox_backend_seed_offset(increment)
            philox_args = torch.tensor(
                [philox_seed, philox_offset], dtype=torch.int64, device=q_device
            )
        else:
            is_dropout = False
            philox_args = torch.empty((2,), dtype=torch.int64, device=q_device)

        p_dropout = 1 - p_dropout
        p_dropout_in_uint8_t = math.floor(p_dropout * 255.0)
        rp_dropout = 1.0 / p_dropout

        if return_softmax:
            assert is_dropout, "Only supported with non-zero dropout."
            p = torch.empty(
                (batch_size, num_heads, seqlen_q_rounded, seqlen_k_rounded),
                device=q_device,
            )
        else:
            p = torch.empty((), device=q_device)

        if zero_tensors:
            out.zero_()
            lse.fill_(float("-inf"))

        params = fwd_params(
            q,  # q_ptr,
            k,  # k_ptr,
            v,  # v_ptr,
            out,  # o_ptr,
            p,  # p_ptr,
            lse,  # softmax_lse_ptr,
            q.stride(-3),  # q_row_stride,
            k.stride(-3),  # k_row_stride,
            v.stride(-3),  # v_row_stride,
            q.stride(-2),  # q_head_stride,
            k.stride(-2),  # k_head_stride,
            v.stride(-2),  # v_head_stride,
            out.stride(-3),  # o_row_stride,
            out.stride(-2),  # o_head_stride,
            q_batch_stride,  # q_batch_stride,
            k_batch_stride,  # k_batch_stride,
            v_batch_stride,  # v_batch_stride,
            o_batch_stride,  # o_batch_stride,
            cu_seqlens_q is not None,  # is_cu_seqlens_q,
            cu_seqlens_q,  # cu_seqlens_q_ptr,
            seqused_k is None,  # is_cu_seqlens_k,
            cu_seqlens_k,  # cu_seqlens_k_ptr,
            seqused_k is not None,  # is_seqused_k,
            seqused_k,  # seqused_k_ptr,
            # sizes
            batch_size,  # b,
            k_batch_size,  # bk,
            num_heads,  # h,
            num_heads_k,  # hk,
            num_heads // num_heads_k,  # h_hk_ratio,
            max_seqlen_q,  # seqlen_q,
            max_seqlen_k,  # seqlen_k,
            seqlen_q_rounded,  # seqlen_q_rounded,
            seqlen_k_rounded,  # seqlen_k_rounded,
            head_size,  # d,
            head_size_rounded,  # d_rounded,
            # scaling factors
            is_softcap,
            adjusted_softcap,  # softcap,
            adjusted_scale_softmax,  # scale_softmax,
            adjusted_scale_softmax_log2e,  # scale_softmax_log2,
            # dropout
            is_dropout,
            p_dropout,
            rp_dropout,
            p_dropout_in_uint8_t,
            philox_args,
            return_softmax,
            # causal and swa
            is_causal,  # is_causal,
            is_local,  # is_local,
            window_size_left,  # window_size_left,
            window_size_right,  # window_size_right,
            seqlenq_ngroups_swapped,  # seqlenq_ngroups_swapped,
            # alibi
            is_alibi,  #
            alibi_slopes,  # alibi_slopes_ptr,
            alibi_slopes_batch_stride,  # alibi_slopes_batch_stride,
            # block table params
            total_q,  # total_q,
            page_table,  # page_table_ptr,
            page_table_batch_stride,  # page_table_batch_stride,
            block_size,  # block_size,
        )

        if flag_gems.vendor_name == "iluvatar":
            params.k_ptr = k.view(k.shape[0], k.shape[1], -1)
            params.v_ptr = v.view(v.shape[0], v.shape[1], -1)
        logger.debug("kernel: flash_varlen_fwd")
        grid = lambda args: (
            triton.cdiv(max_seqlen_q, args["BLOCK_M"]),
            batch_size,
            num_heads,
        )
        kernel = flash_varlen_fwd_kernel[grid]
        args = tuple(getattr(params, k) for k in params.__slots__)

        # We assess which phase the requests are likely to be in and set the config accordingly.
        total_rows = total_q * num_heads
        num_sms = torch_device_fn.get_device_properties(
            flag_gems.device
        ).multi_processor_count
        avg_rows_per_sm = total_rows / num_sms
        avg_rows_per_batch = total_q / batch_size
        avg_rows_per_cta = min(avg_rows_per_batch, avg_rows_per_sm)
        # Heuristic: if avg_rows_per_sm >= 128, we are likely in prefill phase.
        # This is a rough heuristic and may not be accurate for all scenarios.
        if avg_rows_per_cta > 64:
            varlen_fwd_config_str = "mha_block_128"
        elif avg_rows_per_cta > 32:
            varlen_fwd_config_str = "mha_block_64"
        elif avg_rows_per_cta > 16:
            varlen_fwd_config_str = "mha_block_32"
        else:
            varlen_fwd_config_str = "mha_block_16"
        if flag_gems.vendor_name == "mthreads":
            varlen_fwd_config_str = "mha_block_32"

        cfg = runtime.get_heuristic_config(varlen_fwd_config_str)
        cfg_params = {
            "BLOCK_M": cfg["BLOCK_M"](args),
            "BLOCK_N": cfg["BLOCK_N"](args),
            "BLOCK_K": triton.next_power_of_2(head_size),
            "num_warps": cfg["num_warps"](args),
            "num_stages": cfg["num_stages"](args),
        }

        logger.debug("Running flash_varlen_fwd_kernel with config: %s", cfg_params)
        kernel(*args, **cfg_params)

        if seqlenq_ngroups_swapped:
            out = out.reshape(
                batch_size, max_seqlen_q, num_heads_k, head_size
            ).transpose(1, 2)
            if out_ is not None:
                out_.view(batch_size, num_heads_k, max_seqlen_q, head_size).copy_(out)
                out = out_
            else:
                out = out.reshape(batch_size, num_heads_k * max_seqlen_q, head_size)
            lse = lse.reshape(num_heads_k, batch_size, max_seqlen_q)
            lse = lse.reshape(num_heads_k * max_seqlen_q, batch_size)

        unused = torch.empty((), dtype=torch.int64, device=q_device)
    return out, q, k, v, lse, philox_args, unused, p


def mha_fwd(
    q,
    k,
    v,
    out,
    alibi_slopes,
    p_dropout,
    softmax_scale,
    is_causal,
    window_size_left,
    window_size_right,
    softcap,
    return_softmax,
    disable_splitkv=False,
):
    CHECK_DEVICE(q), CHECK_DEVICE(k), CHECK_DEVICE(v)
    q_dtype = q.dtype
    q_device = q.device
    assert q_dtype in (
        torch.float16,
        torch.bfloat16,
    ), "FlashAttention only support fp16 and bf16 data type"
    assert q_dtype == k.dtype
    assert q_dtype == v.dtype
    assert q.stride(-1) == 1, "Input tensor must have contiguous last dimension"
    assert k.stride(-1) == 1, "Input tensor must have contiguous last dimension"
    assert v.stride(-1) == 1, "Input tensor must have contiguous last dimension"
    batch_size, seqlen_q, num_heads, head_size = q.size()
    _, seqlen_k, num_heads_k, _ = k.size()

    # Check output shape
    if out is not None:
        assert out.stride(-1) == 1
        assert out.dtype == q.dtype
        assert out.size() == (batch_size, seqlen_q, num_heads, head_size)
        CHECK_DEVICE(out)

    assert (
        head_size % 8 == 0
    ), "head_size must be a multiple of 8, this is ensured by padding!"
    assert (
        num_heads % num_heads_k == 0
    ), "Number of heads in key/value must divide number of heads in query"
    if window_size_left >= seqlen_k:
        window_size_left = -1
    if window_size_right >= seqlen_k:
        window_size_right = -1
    if seqlen_q == 1 and alibi_slopes is None:
        is_causal = False
    if is_causal:
        window_size_right = 0

    is_causal = window_size_left < 0 and window_size_right == 0
    is_local = window_size_left >= 0 and window_size_right >= 0

    seqlenq_ngroups_swapped = (
        seqlen_q == 1
        and alibi_slopes is None
        and num_heads > num_heads_k
        and window_size_left < 0
        and window_size_right < 0
        and p_dropout == 0
    )
    q_groups = num_heads // num_heads_k

    if seqlenq_ngroups_swapped:
        logger.debug("q_kg swapped.")
        q = q.reshape(batch_size, num_heads_k, q_groups, head_size).transpose(1, 2)
        seqlen_q = q_groups
        num_heads = num_heads_k

    round_multiple = lambda x, m: (x + m - 1) // m * m
    head_size_rounded = round_multiple(head_size, 32)
    seqlen_q_rounded = round_multiple(seqlen_q, 128)
    seqlen_k_rounded = round_multiple(seqlen_k, 32)

    assert (
        head_size <= 256
    ), "FlashAttention forward only supports head dimension at most 256"
    assert head_size == head_size_rounded, "head_size must be rounded to 32"

    def splits_heuristic(num_tasks, num_sms, n_blocks):
        # splits when wave efficiency is low
        n_waves = triton.cdiv(num_tasks, num_sms)
        eff = (num_tasks / num_sms) / n_waves
        if eff > 0.8 or n_waves > 1:
            return 1

        min_blocks_per_split = 2
        best_splits = min(
            triton.cdiv(n_blocks, min_blocks_per_split),
            int(math.floor(1.0 / eff)),
            num_sms,
        )

        return best_splits

    with torch_device_fn.device(q_device):
        # Set softmax params
        lse = torch.empty(
            (batch_size, num_heads, seqlen_q), dtype=torch.float, device=q_device
        )

        if out is not None:
            if seqlenq_ngroups_swapped:
                out = out.reshape(
                    batch_size, num_heads_k, q_groups, head_size
                ).transpose(1, 2)
        else:
            out = torch.empty_like(q, dtype=v.dtype)

        # Set dropout params
        if p_dropout > 0:
            is_dropout = True
            increment = batch_size * num_heads * 32
            philox_seed, philox_offset = philox_backend_seed_offset(increment)
            philox_args = torch.tensor(
                [philox_seed, philox_offset], dtype=torch.int64, device=q_device
            )
        else:
            is_dropout = False
            philox_args = torch.empty((2,), dtype=torch.int64, device=q_device)

        p_dropout = 1 - p_dropout
        p_dropout_in_uint8_t = math.floor(p_dropout * 255.0)
        rp_dropout = 1.0 / p_dropout

        if return_softmax:
            assert is_dropout, "Only supported with non-zero dropout."
            p = torch.empty(
                (batch_size, num_heads, seqlen_q_rounded, seqlen_k_rounded),
                device=q_device,
            )
        else:
            p = torch.empty((), device=q_device)

        M_LOG2E = 1.4426950408889634074
        if softcap > 0.0:
            is_softcap = True
            adjusted_scale_softmax = softcap
            adjusted_softcap = softmax_scale / softcap
            adjusted_scale_softmax_log2e = softcap * M_LOG2E
        else:
            is_softcap = False
            adjusted_softcap = 0.0
            adjusted_scale_softmax = softmax_scale
            adjusted_scale_softmax_log2e = softmax_scale * M_LOG2E

        # Set alibi params
        if alibi_slopes is not None:
            assert alibi_slopes.device == q_device
            assert alibi_slopes.dtype in (torch.float,)
            assert alibi_slopes.stride(-1) == 1
            assert alibi_slopes.shape == (num_heads,) or alibi_slopes.shape == (
                batch_size,
                num_heads,
            )
            alibi_slopes_batch_stride = (
                alibi_slopes.stride(0) if alibi_slopes.ndim == 2 else 0
            )
            is_alibi = True
        else:
            alibi_slopes_batch_stride = 0
            is_alibi = False

        # ONLY EVEN_K IS SUPPORTED
        assert head_size == head_size_rounded

        # Do kernel dispatching
        def dispatch(B, H, Q, K, D, params):
            num_sms = torch_device_fn.get_device_properties(
                "cuda"
            ).multi_processor_count

            # Try bh parallel
            # if B * H > 0.8 * num_sms:
            #     kernel = flash_fwd_bh_parallel_kernel[(H, B)]
            #     # Yield kernel and prefilled args
            #     return kernel, default_args, None, None

            # Try splitkv
            if not is_dropout and not is_local and not disable_splitkv:
                BM = block_m_splitkv_heuristic(D)
                n_tasks = B * H * triton.cdiv(seqlen_q, BM)
                BN = block_n_splitkv_heuristic(D)
                n_blocks = triton.cdiv(seqlen_k, BN)
                n_splits = splits_heuristic(n_tasks, num_sms, n_blocks)

                if n_splits > 1:
                    logger.debug("kernel: flash_fwd_splitkv")
                    lse_splits = torch.empty(
                        (n_splits, B, H, Q), dtype=torch.float, device=q_device
                    )
                    out_splits = torch.empty(
                        (n_splits, B, H, Q, D), dtype=torch.float, device=q_device
                    )
                    grid = lambda args: (
                        triton.cdiv(Q, args["BLOCK_M"]),
                        n_splits,
                        B * H,
                    )
                    splitkv_kernel = flash_fwd_splitkv_kernel[grid]
                    params.o_ptr = out_splits
                    params.softmax_lse_ptr = lse_splits
                    extra_args = {"blocks_per_split": triton.cdiv(n_blocks, n_splits)}
                    kernel = splitkv_kernel(*params.args(), **extra_args)

                    if D >= 128:
                        BLOCK_M = 4
                    elif D >= 64:
                        BLOCK_M = 8
                    else:
                        BLOCK_M = 16
                    BLOCK_K = triton.next_power_of_2(D)
                    grid = lambda args: (triton.cdiv(B * H * Q, BLOCK_M),)
                    combine_kernel = flash_fwd_splitkv_combine_kernel[grid]
                    combine_args = {
                        "out_ptr": out,
                        "lse_ptr": lse,
                        "head_size": head_size,
                        "out_split_stride": out_splits.stride(0),
                        "lse_split_stride": lse_splits.stride(0),
                        "out_b_stride": out.stride(0),
                        "out_s_stride": out.stride(-3),
                        "out_h_stride": out.stride(-1),
                        "out_splits_ptr": out_splits,
                        "lse_splits_ptr": lse_splits,
                        "n_splits": n_splits,
                        "BLOCK_M": BLOCK_M,
                        "BLOCK_K": BLOCK_K,
                        "q_total": B * H * Q,
                        "MAX_N_SPLITS": triton.next_power_of_2(n_splits),
                    }
                    combine_kernel(**combine_args)
                    return kernel

            # Last option: flash_fwd
            logger.debug("kernel: flash_fwd")
            grid = lambda args: (
                triton.cdiv(Q, args["BLOCK_M"]),
                H * B,
            )
            kernel = flash_fwd_kernel[grid]
            kernel = kernel(*params.args())
            return kernel

        if _debug:
            p = torch.empty(
                (batch_size, num_heads, seqlen_q_rounded, seqlen_k_rounded),
                dtype=torch.float32,
                device=q_device,
            )
            return_softmax = True

        params = fwd_params(
            q,  # q_ptr,
            k,  # k_ptr,
            v,  # v_ptr,
            out,  # o_ptr,
            p,  # p_ptr,
            lse,  # softmax_lse_ptr,
            q.stride(-3),  # q_row_stride,
            k.stride(-3),  # k_row_stride,
            v.stride(-3),  # v_row_stride,
            q.stride(-2),  # q_head_stride,
            k.stride(-2),  # k_head_stride,
            v.stride(-2),  # v_head_stride,
            out.stride(-3),  # o_row_stride,
            out.stride(-2),  # o_head_stride,
            q.stride(0),  # q_batch_stride,
            k.stride(0),  # k_batch_stride,
            v.stride(0),  # v_batch_stride,
            out.stride(0),  # o_batch_stride,
            False,  # is_cu_seqlens_q,
            None,  # cu_seqlens_q_ptr,
            False,  # is_cu_seqlens_k,
            None,  # cu_seqlens_k_ptr,
            False,  # is_seqused_k,
            None,  # seqused_k_ptr,
            # sizes
            batch_size,  # b,
            0,  # bk,
            num_heads,  # h,
            num_heads_k,  # hk,
            num_heads // num_heads_k,  # h_hk_ratio,
            seqlen_q,  # seqlen_q,
            seqlen_k,  # seqlen_k,
            seqlen_q_rounded,  # seqlen_q_rounded,
            seqlen_k_rounded,  # seqlen_k_rounded,
            head_size,  # d,
            head_size_rounded,  # d_rounded,
            # scaling factors
            is_softcap,
            adjusted_softcap,  # softcap,
            adjusted_scale_softmax,  # scale_softmax,
            adjusted_scale_softmax_log2e,  # scale_softmax_log2,
            # dropout
            is_dropout,
            p_dropout,
            rp_dropout,
            p_dropout_in_uint8_t,
            philox_args,
            return_softmax,
            # causal and swa
            is_causal,  # is_causal,
            is_local,  # is_local,
            window_size_left,  # window_size_left,
            window_size_right,  # window_size_right,
            seqlenq_ngroups_swapped,  # seqlenq_ngroups_swapped,
            # alibi
            is_alibi,  #
            alibi_slopes,  # alibi_slopes_ptr,
            alibi_slopes_batch_stride,  # alibi_slopes_batch_stride,
            # block table params
            0,  # total_q,
            None,  # page_table_ptr,
            0,  # page_table_batch_stride,
            0,  # block_size,
        )

        # Move TxD to last dims for correct stride in Triton tt.load
        if flag_gems.vendor_name == "iluvatar":
            params.q_ptr = q.transpose(1, 2)
            params.k_ptr = k.transpose(1, 2)
            params.v_ptr = v.transpose(1, 2)
        kernel = dispatch(batch_size, num_heads, seqlen_q, seqlen_k, head_size, params)

        if _debug:
            print(f"{kernel.name} shared memory:", kernel.metadata.shared)
            print(f"{kernel.name} num_warps:", kernel.metadata.num_warps)
            print(f"{kernel.name} num_stages:", kernel.metadata.num_stages)
            # print(kernel.asm['ttgir'])

        if seqlenq_ngroups_swapped:
            out = out.transpose(1, 2).reshape(
                (batch_size, 1, num_heads_k * seqlen_q, head_size)
            )
            q = q.transpose(1, 2).reshape(
                (batch_size, 1, num_heads_k * seqlen_q, head_size)
            )
            lse = lse.reshape((batch_size, num_heads_k * seqlen_q, 1))

        unused = torch.empty((), dtype=torch.int64, device=q_device)

    return out, q, k, v, lse, philox_args, unused, p



"""
Universal Neg - Triton Optimized
"""
import torch
import triton
import triton.language as tl

@triton.jit
def _neg_kernel(x_ptr, out_ptr, n_elements, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    
    x = tl.load(x_ptr + offsets, mask=mask)
    output = -x
    tl.store(out_ptr + offsets, output, mask=mask)

def _unary_impl(x: torch.Tensor) -> torch.Tensor:
    """Unary op implementation using pure Triton."""
    assert x.is_cuda, "Tensor must be on CUDA"
    x = x.contiguous()
    
    output = torch.empty_like(x)
    n_elements = output.numel()
    
    if n_elements == 0:
        return output
    
    BLOCK_SIZE = 1024
    grid = (triton.cdiv(n_elements, BLOCK_SIZE),)
    
    _neg_kernel[grid](
        x, output,
        n_elements,
        BLOCK_SIZE=BLOCK_SIZE,
    )
    
    return output

def neg_kernel(x: torch.Tensor) -> torch.Tensor:
    """
    Neg activation using pure Triton.
    
    Args:
        x: Input tensor
    
    Returns:
        Result tensor
    """
    return _unary_impl(x)



import importlib
import logging
import math
import os
from typing import Callable, Mapping

import torch

from flag_gems.utils.code_cache import cache_dir
from flag_gems.utils.code_utils import IndentedBuffer, write_atomic

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


class VstackKernelCode(IndentedBuffer):
    """
    Vstack kernel template.
    """

    overloads: Mapping[str, Callable] = {}

    def __init__(self):
        self.pid = os.getpid()
        self.cache = self.overloads
        self.kernel_name = "_vstack_jit_kernel"
        self.wrapper_func_name = "_wrapper"
        self.vstack_small_limit = 49152
        super(VstackKernelCode, self).__init__()

    def __init(self, tensors):
        """Initialize the vstack kernel."""
        self.device = tensors[0].device
        self.dtype = tensors[0].dtype
        for tensor in tensors:
            assert (
                tensor.device == self.device
                and tensor.dtype == self.dtype
                and tensors[0].shape[1:] == tensor.shape[1:]
            )
        c_tensors = [t.contiguous() for t in tensors]
        self.inputs = []
        self.idxs = [0]
        self.total_size = 0
        for tensor in c_tensors:
            self.total_size += tensor.numel()
            self.idxs.append(self.total_size)
            self.inputs.append(tensor)
        self.deal_num = math.ceil(self.total_size / TOTAL_CORE_NUM)
        self.input_num = len(self.inputs)
        flag = (self.total_size / self.input_num) == self.idxs[1]
        if (
            self.total_size < self.vstack_small_limit
            and self.input_num <= TOTAL_CORE_NUM
            and flag
        ):
            self.is_small = True
        else:
            self.is_small = False

    def __imports(self):
        """Generate imports for the kernel code."""
        self.tpl(
            """
import math
import torch
import triton
from triton import language as tl
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, libtuner
from flag_gems.runtime.backend import vendor_module
TOTAL_CORE_NUM = vendor_module.TOTAL_CORE_NUM
MAX_NRAM_SIZE = vendor_module.MAX_NRAM_SIZE
        """
        )

    def __wrapper(self):
        """Generate wrapper function for the kernel code."""
        self.newline()
        self.tpl(
            """
def {wrapper_name}(tensors, inputs, idx, total_size, input_num, deal_num, is_small):
    tensors = torch.atleast_2d(tensors)
    num_tensors = len(tensors)
    assert num_tensors > 0
    if num_tensors == 1:
        return tensors[0]
    device = tensors[0].device
    dtype = tensors[0].dtype
    input = [i for i in inputs]
    c_tensors = [t.contiguous() for t in tensors]
    total_rows = sum(tensor.shape[0] for tensor in c_tensors)
    output_shape = list(c_tensors[0].shape)
    output_shape[0] = total_rows
    output = torch.empty(output_shape, device=device, dtype=dtype)
    with torch_device_fn.device(device):
        {kernel_name}[(TOTAL_CORE_NUM,)]({args})
    return output
        """,
            wrapper_name=self.wrapper_func_name,
            kernel_name=self.kernel_name,
            args=self.__kernel_args(is_declare=False),
        )

    def __config(self):
        """Generate config for the kernel code."""
        # generate config key.
        self.newline()
        self.tpl(
            """
@libentry()
@libtuner(
    configs=[
        triton.Config({{'BLOCK_SIZE' : 512}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 2048}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 4096}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 8192}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 10240}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 14336}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 18000}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 22000}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 28000}}, num_warps=1),
        triton.Config({{'BLOCK_SIZE' : 32000}}, num_warps=1),
    ],
    key = [{config_keys}],
)
@triton.jit
        """,
            config_keys="'total_size'",
        )

    def __kernel(self):
        """Generate kernel code body."""
        # configuration.
        self.__config()
        kernel_signature = f"def {self.kernel_name}({self.__kernel_args()}):"
        self.idx_1 = 1
        self.idx_0 = 0
        self.writeline(kernel_signature)
        with self.indent():
            self.writeline("pid_x = tl.program_id(0)")
            self.writeline("block = tl.arange(0, BLOCK_SIZE)")
            self.writeline("if is_small:")
            with self.indent():
                self.writeline("for i in range(input_num):")
                with self.indent():
                    for i in range(self.input_num):
                        self.writeline(f"if pid_x == {i} and pid_x == i:")
                        with self.indent():
                            self.writeline(
                                f"for num in range(0, idx_{i + 1} - idx_{i}, BLOCK_SIZE):"
                            )
                            with self.indent():
                                self.writeline("in_offset = num + block")
                                self.writeline(f"dst_offset = idx_{i} + num + block")
                                self.writeline(
                                    f"x = tl.load(input_{i} + in_offset, mask = in_offset < idx_{i + 1} - idx_{i})"
                                )
                                self.writeline(
                                    f"tl.store(output + dst_offset, x, mask = dst_offset < idx_{i + 1})"
                                )
            self.writeline("else:")
            with self.indent():
                self.writeline("condidate_num = idx_1")
                self.writeline("input_iter = 0")
                self.writeline("for pid in range(pid_x + 1):")
                with self.indent():
                    self.writeline("need_num = deal_num")
                    self.writeline("while(need_num > 0):")
                    with self.indent():
                        self.writeline("per_fetch_num = min(condidate_num, need_num)")
                        self.writeline("if pid == pid_x:")
                        with self.indent():
                            self.writeline("if input_iter == 0:")
                            with self.indent():
                                self.writeline("offset = idx_1 - idx_0 - condidate_num")
                                self.writeline("deal_rem = deal_num - per_fetch_num")
                                self.writeline(
                                    "for i in range(0, deal_num, BLOCK_SIZE):"
                                )
                                with self.indent():
                                    self.writeline("in_offset = offset + i + block")
                                    self.writeline("dst_offset = in_offset")
                                    self.writeline(
                                        "x = tl.load(input_0 + in_offset, mask=in_offset < idx_1 - idx_0)"
                                    )
                                    self.writeline(
                                        "tl.store(output + dst_offset, x, mask=dst_offset<idx_1)"
                                    )
                            self.writeline("else:")
                            with self.indent():
                                for i in range(1, self.input_num, 1):
                                    idx = i + 1
                                    self.writeline(f"if input_iter == {i}:")
                                    with self.indent():
                                        self.writeline(
                                            f"offset = idx_{idx} - idx_{i} - condidate_num"
                                        )
                                        self.writeline("if need_num != deal_num:")
                                        with self.indent():
                                            self.writeline(
                                                "deal_rem = deal_num - per_fetch_num"
                                            )
                                            self.writeline(
                                                "for i in range(0, need_num, BLOCK_SIZE):"
                                            )
                                            with self.indent():
                                                self.writeline(
                                                    "in_offset = offset + i + block"
                                                )
                                                self.writeline(
                                                    f"dst_offset = idx_{i} + in_offset"
                                                )
                                                self.writeline(
                                                    f"x = tl.load(input_{i} + in_offset, mask=in_offset < need_num)"
                                                )
                                                self.writeline(
                                                    f"tl.store(output + dst_offset, x, \
                                                        mask=dst_offset<idx_{i}+per_fetch_num)"
                                                )
                                        self.writeline("else:")
                                        with self.indent():
                                            self.writeline(
                                                "for i in range(0, need_num, BLOCK_SIZE):"
                                            )
                                            with self.indent():
                                                self.writeline(
                                                    "in_offset = offset + i + block"
                                                )
                                                self.writeline(
                                                    f"dst_offset = idx_{i} + in_offset"
                                                )
                                                self.writeline(
                                                    f"x = tl.load(input_{i} + in_offset, \
                                                        mask=in_offset < idx_{idx}-idx_{i})"
                                                )
                                                self.writeline(
                                                    f"tl.store(output + dst_offset, x, mask=dst_offset<idx_{idx})"
                                                )
                        self.writeline("condidate_num -= per_fetch_num")
                        self.writeline("need_num -= per_fetch_num")
                        self.writeline("if (condidate_num <= 0):")
                        with self.indent():
                            for i in range(1, self.input_num, 1):
                                idx = i + 1
                                input_idx = i - 1
                                if self.input_num == 2:
                                    self.writeline(
                                        f"condidate_num = idx_{idx} - idx_{i}"
                                    )
                                else:
                                    if i == 1:
                                        self.writeline(f"if input_iter == {input_idx}:")
                                        with self.indent():
                                            self.writeline(
                                                f"condidate_num = idx_{idx} - idx_{i}"
                                            )
                                    else:
                                        if i < self.input_num - 1:
                                            self.writeline(
                                                f"elif input_iter == {input_idx}:"
                                            )
                                            with self.indent():
                                                self.writeline(
                                                    f"condidate_num = idx_{idx} - idx_{i}"
                                                )
                                        else:
                                            self.writeline("else:")
                                            with self.indent():
                                                self.writeline(
                                                    f"condidate_num = idx_{idx} - idx_{i}"
                                                )

                            self.writeline("input_iter += 1")

    def __gen_code(self):
        """Entry point for code generation of vstack."""
        # generate imports.
        self.__imports()
        # generate wrapper function.
        self.__wrapper()

        # generate kernel.
        self.__kernel()

    def __kernel_args(self, is_declare=True):
        input_args = []
        idxs_args = []
        if is_declare:
            for i in range(self.input_num):
                input_args.append(f"input_{i}")
            for i in range(len(self.idxs)):
                idxs_args.append(f"idx_{i}")
        else:
            for i in range(self.input_num):
                input_args.append(f"input[{i}]")
            for i in range(len(self.idxs)):
                idxs_args.append(f"idx[{i}]")
        input_args_str = ", ".join(input_args)
        idxs_args_str = ", ".join(idxs_args)

        extra_args_str = f"{input_args_str}, {idxs_args_str}"
        if is_declare:
            return f"{extra_args_str}, output, total_size, input_num, deal_num, is_small, BLOCK_SIZE: tl.constexpr"
        else:
            return (
                f"{extra_args_str}, output, total_size, input_num, deal_num, is_small"
            )

    def __call__(self, tensors: list) -> torch.Tensor:
        # get overload kernel.
        self.__init(tensors)

        vstack_input_num = "_".join(str(self.input_num))

        self.kernel_name = self.kernel_name + "_vstack_" + vstack_input_num
        key = f"{self.total_size}_{self.input_num}"
        if key not in self.cache:
            # generate code and cache.
            self.__gen_code()
            file_name = f"vstack_{key}_pid_{self.pid}.py"
            filepath = cache_dir() / file_name
            write_atomic(filepath, self.getvalue())
            # load
            spec = importlib.util.spec_from_file_location(
                f"_gen_module_{key}_pid_{self.pid}", filepath
            )
            m = importlib.util.module_from_spec(spec)
            # do not expose it to sys.modules
            # sys.modules["_add_module"] = m
            spec.loader.exec_module(m)
            overload = getattr(m, self.wrapper_func_name)
            self.cache[key] = overload
        overload = self.cache[key]
        return overload(
            tensors,
            self.inputs,
            self.idxs,
            self.total_size,
            self.input_num,
            self.deal_num,
            self.is_small,
        )


def vstack(tensors: list):
    logger.debug("GEMS_CAMBRICON VSTACK")

    return VstackKernelCode()(tensors)



import logging

import triton
import triton.language as tl

from flag_gems.utils import tl_extra_shim

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
exp2 = tl_extra_shim.exp2


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def sigmoid_forward(x):
    # log2e: tl.constexpr = math.log2(math.e)
    # triton 3.0.0 disallow calling non-jitted function inside jitted function, even if it is in
    # the rhs of an assignment to a constexpr, so we use numeric literal instead to work around this.
    # log2e: tl.constexpr = 1.4426950408889634
    return 1 / (1 + tl.exp(-x.to(tl.float32)))


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def sigmoid_backward_kernel(dy, y):
    y_f32 = y.to(tl.float32)
    dy_f32 = dy.to(tl.float32)
    return dy_f32 * (1.0 - y_f32) * y_f32


def sigmoid(self):
    logger.debug("GEMS SIGMOID FORWARD")
    output = sigmoid_forward(self)
    return output


def sigmoid_backward(grad_output, output):
    logger.debug("GEMS SIGMOID BACKWARD")
    grad_input = sigmoid_backward_kernel(grad_output, output)
    return grad_input


def sigmoid_(A):
    logger.debug("GEMS SIGMOID_ FORWARD")
    out = sigmoid_forward(A, out0=A)
    return out



"""
MoE Fusion Pass — DAG Rewrite for Mixture-of-Experts Models

Detects MoE routing subgraphs (topk → sort → bincount → per-expert FFN → scatter_reduce)
and replaces them with a single fused moe_dispatch op that executes dynamically at runtime.

This eliminates hardcoded slice boundaries burned into the graph during tracing,
enabling correct routing for any input at runtime.

ZERO HARDCODE: All parameters (num_experts, top_k, hidden_dim, intermediate_dim)
are extracted from the graph tensors and op attributes.

Called BEFORE CompiledSequence.compile() to transform the DAG in-place.
"""

from typing import Any, Dict, List, Optional, Set, Tuple
import re
import os


def detect_and_fuse_moe(dag: Dict[str, Any], family: str, norm_topk_prob: bool = True) -> Dict[str, Any]:
    """
    Detect MoE patterns in DAG and replace with fused ops.

    Guards:
        - family must be "llm" (eliminates image/audio/video)
        - Must find topk ops with k > 1 (eliminates dense LLMs)

    Args:
        dag: The TensorDAG dict (mutated in-place)
        family: Model family from manifest ("llm", "image", etc.)
        norm_topk_prob: Whether to normalize routing weights after topk selection.
            DeepSeek: False (raw softmax scores). Qwen3/Mixtral: True (default).

    Returns:
        The DAG (same reference, possibly mutated)
    """
    if os.environ.get("NBX_DISABLE_MOE_FUSION"):
        return dag
    if family != "llm":
        return dag

    ops = dag.get("ops", {})
    execution_order = dag.get("execution_order", [])
    tensors = dag.get("tensors", {})

    # Single O(N) pass: find all topk ops with k > 1
    moe_topk_uids = []
    for op_uid in execution_order:
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        if op_data.get("op_type") != "aten::topk":
            continue
        # Extract k from attributes.args[1] (positional arg)
        k = _extract_topk_k(op_data)
        if k is not None and k > 1:
            moe_topk_uids.append(op_uid)

    if not moe_topk_uids:
        return dag  # Not MoE — unchanged

    # Build maps ONCE, then update incrementally after each fusion
    consumer_map = _build_consumer_map(ops, execution_order)
    producer_map = _build_producer_map(ops, execution_order)

    fused_count = 0
    ops_removed_total = 0

    for topk_uid in moe_topk_uids:
        result = _fuse_one_moe_layer(
            dag, ops, execution_order, tensors,
            consumer_map, producer_map, topk_uid,
            norm_topk_prob=norm_topk_prob,
        )
        if result is not None:
            removed_count, fused_uid, fused_op = result
            fused_count += 1
            ops_removed_total += removed_count

            # Incremental map update: add fused op's inputs/outputs
            for in_tid in _collect_input_tids(fused_op):
                if in_tid not in consumer_map:
                    consumer_map[in_tid] = []
                consumer_map[in_tid].append(fused_uid)
            for out_tid in fused_op.get("output_tensor_ids", []):
                producer_map[out_tid] = fused_uid

    # Update DAG
    dag["ops"] = ops
    dag["execution_order"] = execution_order

    return dag


def _fuse_one_moe_layer(
    dag: Dict[str, Any],
    ops: Dict[str, Any],
    execution_order: List[str],
    tensors: Dict[str, Any],
    consumer_map: Dict[str, List[str]],
    producer_map: Dict[str, str],
    topk_uid: str,
    norm_topk_prob: bool = True,
) -> Optional[Tuple[int, str, Dict[str, Any]]]:
    """
    Fuse one MoE layer starting from a topk op.

    Returns (ops_removed, fused_uid, fused_op_data) or None if fusion failed/skipped.

    COMPATIBILITY:
    - DeepSeek v1 (deepseek-moe-16b-chat): ~1300 ops per layer, mm ops in subgraph → FUSE
    - DeepSeek v2 (DeepSeek-Coder-V2): ~11 ops per layer, mm ops NOT in subgraph → SKIP
      V2 uses a different architecture where expert computation is decoupled from routing.
    """
    topk_data = ops[topk_uid]

    # --- Extract MoE parameters from graph (ZERO HARDCODE) ---

    # topk input = gate_scores (softmax output)
    gate_scores_tid = _get_input_tensor_id(topk_data, 0)
    if gate_scores_tid is None:
        raise RuntimeError(f"[MoE Fusion] Cannot find gate_scores input for {topk_uid}")

    # k from topk attributes
    top_k = _extract_topk_k(topk_data)

    # topk outputs: scores and indices
    topk_scores_tid = topk_data["output_tensor_ids"][0]
    topk_indices_tid = topk_data["output_tensor_ids"][1]

    # --- Trace forward from topk to find the complete MoE subgraph ---

    # Collect ALL ops that belong to this MoE layer
    moe_op_uids: Set[str] = set()
    moe_op_uids.add(topk_uid)

    # Find the hidden_states input: trace back from the index ops
    # The gate_scores come from softmax, which comes from router mm
    # The hidden_states are the input to both the router AND the expert index ops

    # Step 1: Find sort, bincount, floor_divide, view ops after topk
    _trace_routing_ops(ops, execution_order, consumer_map,
                       topk_scores_tid, topk_indices_tid, moe_op_uids)

    # num_experts from topk input shape (gate_scores last dim) — ZERO HARDCODE
    num_experts = _count_total_experts(tensors, topk_uid, ops)

    # Step 2: Extract expert weight IDs BEFORE removing boundary ops
    # Boundary removal would exclude mm ops whose outputs exit the subgraph (down projection),
    # but we need those mm ops to extract the expert weight tensor IDs.
    expert_weight_ids, num_experts_found, hidden_states_tid = \
        _trace_expert_blocks(ops, execution_order, tensors, consumer_map,
                             producer_map, moe_op_uids, topk_uid, num_experts)

    # Step 3: Remove boundary ops — ops whose outputs have consumers OUTSIDE the MoE subgraph.
    # These ops must stay in execution_order. This happens AFTER weight extraction.
    boundary_ops = set()
    for op_uid in list(moe_op_uids):
        op_data = ops.get(op_uid, {})
        for out_tid in op_data.get("output_tensor_ids", []):
            for c_uid in consumer_map.get(out_tid, []):
                if c_uid not in moe_op_uids:
                    boundary_ops.add(op_uid)
                    break
            if op_uid in boundary_ops:
                break
    moe_op_uids -= boundary_ops

    # ═══════════════════════════════════════════════════════════════════════════
    # COMPATIBILITY CHECK: Skip fusion if expert weights not found in MoE subgraph
    # ═══════════════════════════════════════════════════════════════════════════
    # DeepSeek v2 architecture decouples routing from expert execution.
    # The mm ops with expert weights are NOT in the topk-derived subgraph.
    # In this case, we SKIP fusion — the model runs correctly without it.
    if num_experts_found == 0:
        return None

    # Step 3: Find MoE output tensor — DATA-DRIVEN (works for v1 scatter_reduce AND v2 index_put+sum)
    last_scatter_tid = _find_moe_output(
        ops, execution_order, tensors, consumer_map, producer_map,
        moe_op_uids, boundary_ops
    )

    if last_scatter_tid is None:
        return None

    if hidden_states_tid is None:
        return None

    # --- Create fused op ---
    # Extract block identifier from parent_module of topk
    parent = topk_data.get("parent_module", "")
    # e.g. "block.1.ffn.router" -> "block.1"
    block_match = re.match(r"(block\.\d+)", parent)
    block_id = block_match.group(1) if block_match else topk_uid

    fused_uid = f"moe_fused::{block_id}"

    # Build args list for liveness analysis — all input tensors must be declared
    # so _extract_input_slots_from_dag can track them and prevent premature GC
    liveness_args = [
        {"type": "tensor", "tensor_id": hidden_states_tid},
        {"type": "tensor", "tensor_id": gate_scores_tid},
    ]
    # All expert weight tensors (64 × 3 = 192 tensors)
    for i in range(num_experts):
        liveness_args.append({"type": "tensor", "tensor_id": expert_weight_ids["gate"][i]})
        liveness_args.append({"type": "tensor", "tensor_id": expert_weight_ids["up"][i]})
        liveness_args.append({"type": "tensor", "tensor_id": expert_weight_ids["down"][i]})

    # input_tensor_ids for native mode liveness tracking (_compute_last_use scans this)
    all_input_tids = [hidden_states_tid, gate_scores_tid]
    for i in range(num_experts):
        all_input_tids.append(expert_weight_ids["gate"][i])
        all_input_tids.append(expert_weight_ids["up"][i])
        all_input_tids.append(expert_weight_ids["down"][i])

    fused_op = {
        "op_type": "custom::moe_fused",
        "output_tensor_ids": [last_scatter_tid],
        "input_tensor_ids": all_input_tids,
        "output_shapes": tensors.get(last_scatter_tid, {}).get("shape", []),
        "attributes": {
            "args": liveness_args,
            "kwargs": {},
            "gate_scores_tid": gate_scores_tid,
            "hidden_states_tid": hidden_states_tid,
            "expert_gate_weight_ids": expert_weight_ids["gate"],
            "expert_up_weight_ids": expert_weight_ids["up"],
            "expert_down_weight_ids": expert_weight_ids["down"],
            "top_k": top_k,
            "num_experts": num_experts,
            "norm_topk_prob": norm_topk_prob,
        },
    }

    # --- Remove old ops from execution_order, add fused op ---

    # Find the latest non-MoE op position that is BEFORE or WITHIN the MoE range.
    # The fused op must come AFTER all its input producers (hidden_states, gate_scores).
    # These producers may be interleaved with MoE ops in execution_order.

    # Find the position of the LAST MoE op in the original order
    last_moe_pos = 0
    for i, uid in enumerate(execution_order):
        if uid in moe_op_uids:
            last_moe_pos = i

    # Remove all MoE ops and build new order
    new_order = []
    for uid in execution_order:
        if uid in moe_op_uids:
            continue
        new_order.append(uid)

    # Insert fused op right after the last non-MoE op that was before or within
    # the MoE block range. This ensures all input producers have executed.
    insert_idx = 0
    for uid in execution_order[:last_moe_pos + 1]:
        if uid not in moe_op_uids:
            insert_idx += 1
    new_order.insert(insert_idx, fused_uid)

    # Post-fusion dead-op elimination: boundary ops that depend on removed
    # internal MoE ops will get None inputs → remove them.
    # The fused op produces last_scatter_tid; any other internal tensor is gone.
    #
    # PROTECTION: Never remove shared_expert ops — they have consumers outside
    # the MoE subgraph (residual connections). The parent_module heuristic
    # catches "shared_expert" or "shared" in the module path.
    fused_output_tids = set(fused_op.get("output_tensor_ids", []))
    removed_producers = set()
    for uid in moe_op_uids:
        for out_tid in ops.get(uid, {}).get("output_tensor_ids", []):
            if out_tid not in fused_output_tids:
                removed_producers.add(out_tid)

    # Iteratively remove ops whose inputs depend on removed tensors
    dead_ops: Set[str] = set()
    changed = True
    while changed:
        changed = False
        for uid in new_order:
            if uid in dead_ops or uid == fused_uid:
                continue
            # PROTECT shared_expert paths from dead-op elimination
            op_data = ops.get(uid, {})
            parent_module = op_data.get("parent_module", "")
            if "shared_expert" in parent_module or "shared" in parent_module:
                continue
            for in_tid in _collect_input_tids(op_data):
                if in_tid in removed_producers:
                    dead_ops.add(uid)
                    # This dead op's outputs are also removed
                    for out_tid in op_data.get("output_tensor_ids", []):
                        removed_producers.add(out_tid)
                    changed = True
                    break

    if dead_ops:
        new_order = [uid for uid in new_order if uid not in dead_ops]

    # Update structures
    ops[fused_uid] = fused_op
    # Don't delete old ops from ops dict (tensors reference them), just remove from execution_order
    execution_order.clear()
    execution_order.extend(new_order)

    ops_removed = len(moe_op_uids)

    return ops_removed, fused_uid, fused_op


# ============================================================================
# TRACING HELPERS
# ============================================================================

def _trace_routing_ops(
    ops: Dict[str, Any],
    execution_order: List[str],
    consumer_map: Dict[str, List[str]],
    topk_scores_tid: str,
    topk_indices_tid: str,
    moe_op_uids: Set[str],
) -> None:
    """
    Trace routing ops after topk: view, sort, bincount, floor_divide,
    _to_copy, detach, zeros_like, etc.

    Phase 1: Forward BFS from topk outputs to find all MoE ops.
    Phase 2: Backward pass to collect ops whose outputs are ONLY consumed by MoE ops
             (e.g., zeros_like that feeds scatter_reduce).
    """
    # MoE op types — any op with these types that consumes a topk-derived tensor
    # is part of the MoE subgraph
    MOE_OP_TYPES = {
        # Routing infrastructure
        "aten::view", "aten::reshape", "aten::_unsafe_view",
        "aten::sort", "aten::bincount", "aten::floor_divide",
        "aten::_to_copy", "aten::detach",
        "aten::mul", "aten::div", "aten::sum",
        "aten::unsqueeze", "aten::squeeze", "aten::permute",
        # Expert selection (Qwen3-MoE pattern: gt → nonzero → unbind → per-expert paths)
        "aten::gt", "aten::lt", "aten::ge", "aten::le", "aten::eq", "aten::ne",
        "aten::nonzero", "aten::unbind", "aten::_local_scalar_dense",
        # Expert block ops
        "aten::select",   # Qwen3: select(routing_matrix, dim=0, index=expert_id) per expert
        "aten::slice", "aten::index", "aten::t", "aten::mm", "aten::silu",
        "aten::scatter_reduce", "aten::repeat",
        "aten::index_put", "aten::scatter_add", "aten::index_add",
        # V2 aggregation pattern: cat + scatter (NOT aten::add — escapes via residuals)
        "aten::cat", "aten::scatter",
        # Accumulation setup (buffers for expert outputs)
        "aten::zeros_like", "aten::zeros", "aten::full",
        "aten::empty_like", "aten::empty",
    }

    # Phase 1: Forward BFS from topk outputs
    visited_tids: Set[str] = set()
    frontier = [topk_scores_tid, topk_indices_tid]

    while frontier:
        tid = frontier.pop()
        if tid in visited_tids:
            continue
        visited_tids.add(tid)

        for consumer_uid in consumer_map.get(tid, []):
            if consumer_uid in moe_op_uids:
                continue
            consumer_data = ops.get(consumer_uid)
            if consumer_data is None:
                continue

            op_type = consumer_data.get("op_type", "")
            if op_type in MOE_OP_TYPES:
                moe_op_uids.add(consumer_uid)
                for out_tid in consumer_data.get("output_tensor_ids", []):
                    frontier.append(out_tid)



def _find_moe_output(
    ops: Dict[str, Any],
    execution_order: List[str],
    tensors: Dict[str, Any],
    consumer_map: Dict[str, List[str]],
    producer_map: Dict[str, str],
    moe_op_uids: Set[str],
    boundary_ops: Set[str],
) -> Optional[str]:
    """
    Find the MoE output tensor using DATA-DRIVEN detection.

    Instead of hardcoding op_type (scatter_reduce for v1, index_put for v2, etc.),
    we find the last MoE op whose output has consumers OUTSIDE the MoE subgraph.
    This is the definition of "MoE output" — the tensor that continues in the network.

    Strategy 1: Find op in moe_op_uids whose output exits the subgraph
    Strategy 2: If boundary_ops contains the exit op, find it there
    Strategy 3: Fallback to last op with [seq_len, hidden_dim] shaped output

    Returns:
        output tensor_id, or None if not found
    """
    # Strategy 1: Find op whose output exits the MoE subgraph
    for op_uid in reversed(execution_order):
        if op_uid not in moe_op_uids:
            continue
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        for out_tid in op_data.get("output_tensor_ids", []):
            consumers = consumer_map.get(out_tid, [])
            has_external_consumer = any(c not in moe_op_uids for c in consumers)
            if has_external_consumer:
                # Validate: output should be [seq_len, hidden_dim], not a routing scalar
                tdata = tensors.get(out_tid, {})
                shape = tdata.get("shape", [])
                if len(shape) >= 2 and shape[-1] > 1:
                    return out_tid

    # Strategy 2: boundary_ops were removed from moe_op_uids because they have external consumers.
    # One of them IS the MoE output — find the one whose input comes from moe_op_uids.
    for op_uid in reversed(execution_order):
        if op_uid not in boundary_ops:
            continue
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        # Check if this boundary op consumes a tensor from the MoE subgraph
        has_moe_input = False
        for in_tid in _collect_input_tids(op_data):
            producer_uid = producer_map.get(in_tid)
            if producer_uid in moe_op_uids:
                has_moe_input = True
                break
        if has_moe_input:
            out_tids = op_data.get("output_tensor_ids", [])
            if out_tids:
                tdata = tensors.get(out_tids[0], {})
                shape = tdata.get("shape", [])
                if len(shape) >= 2 and shape[-1] > 1:
                    return out_tids[0]

    # Strategy 3: Fallback — last MoE op with activation-sized output
    for op_uid in reversed(execution_order):
        if op_uid not in moe_op_uids:
            continue
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        out_tids = op_data.get("output_tensor_ids", [])
        if out_tids:
            tdata = tensors.get(out_tids[0], {})
            shape = tdata.get("shape", [])
            # Activation-sized: at least 2D with hidden_dim > routing_dim
            # e.g., [64, 2048] not [64, 6] or [64]
            if len(shape) >= 2 and shape[-1] > 64:
                return out_tids[0]

    return None


def _trace_expert_blocks(
    ops: Dict[str, Any],
    _execution_order: List[str],  # unused, kept for API compatibility
    tensors: Dict[str, Any],
    _consumer_map: Dict[str, List[str]],  # unused, kept for API compatibility
    producer_map: Dict[str, str],
    moe_op_uids: Set[str],
    _topk_uid: str,  # unused, kept for API compatibility
    num_experts: int = 0,
) -> Tuple[Dict[str, List[str]], int, Optional[str]]:
    """
    Extract expert weight tensor IDs from MM ops in the MoE subgraph.

    Returns:
        expert_weight_ids: {"gate": [...], "up": [...], "down": [...]} ordered by expert ID
        num_experts_found: number of distinct experts with ops in graph
        hidden_states_tid: tensor ID of the hidden_states input to experts

    NOTE: last_scatter_tid is now found by _find_moe_output() which uses data-driven
    detection instead of hardcoded op_type matching.
    """
    # Pattern: param::block.X.ffn.expert.Y.{gate,up,down}.weight
    # This pattern matches:
    #   - DeepSeek v1/v2: param::block.1.ffn.expert.0.gate.weight
    #   - Mixtral (future): param::model.layers.1.block_sparse_moe.experts.0.gate_proj.weight
    # Group 1 = expert_id, Group 2 = gate|up|down
    weight_pattern = re.compile(
        r"param::.*\.expert[s]?\.(\d+)\.(gate|up|down)(?:_proj)?\.weight"
    )

    # Collect weight tensor IDs from mm ops in the MoE subgraph
    expert_weights: Dict[int, Dict[str, str]] = {}  # expert_id -> {gate: tid, up: tid, down: tid}
    hidden_states_tid = None

    for op_uid in moe_op_uids:
        op_data = ops.get(op_uid)
        if op_data is None:
            continue

        # Find weight references in mm ops
        # mm consumes [activation, transposed_weight]. The transposed_weight comes from
        # an aten::t op whose input is param::block.X.ffn.expert.Y.{gate,up,down}.weight
        if op_data.get("op_type") == "aten::mm":
            # Check both inputs of mm for weight references
            attrs = op_data.get("attributes", {})
            for arg in attrs.get("args", []):
                if arg.get("type") != "tensor":
                    continue
                tid = arg.get("tensor_id", "")
                # The mm input is the t output (e.g. aten.t::12::out_0)
                # Trace back to find the param:: weight
                producer_uid = producer_map.get(tid)
                if producer_uid is None:
                    continue
                producer_data = ops.get(producer_uid, {})
                if producer_data.get("op_type") == "aten::t":
                    # The transpose input is the weight tensor
                    t_input = _get_input_tensor_id(producer_data, 0)
                    if t_input is None:
                        continue
                    m = weight_pattern.match(t_input)
                    if m:
                        expert_id = int(m.group(1))  # Group 1 = expert ID
                        proj_type = m.group(2)       # Group 2 = gate|up|down
                        if expert_id not in expert_weights:
                            expert_weights[expert_id] = {}
                        expert_weights[expert_id][proj_type] = t_input

        # Find hidden_states: input to an index op that gathers expert tokens
        # Must be float (not int64 routing tensors), large hidden_dim,
        # and produced OUTSIDE the MoE subgraph (not an intermediate routing tensor).
        if op_data.get("op_type") == "aten::index" and hidden_states_tid is None:
            input_tid = _get_input_tensor_id(op_data, 0)
            if input_tid is not None and not input_tid.startswith("param::"):
                tdata = tensors.get(input_tid, {})
                shape = tdata.get("shape", [])
                dtype = tdata.get("dtype", "")
                # Hidden states: 2D [seq, hidden] or 3D [batch, seq, hidden]
                # with large last dim (hidden_dim, e.g. 2048), float dtype,
                # and produced by a non-MoE op (not an intermediate routing tensor)
                if (len(shape) in (2, 3) and shape[-1] > 64
                        and "int" not in dtype
                        and producer_map.get(input_tid) not in moe_op_uids):
                    hidden_states_tid = input_tid

    # Build ordered weight ID lists (0..num_experts-1)
    num_experts_found = len(expert_weights)
    gate_ids = []
    up_ids = []
    down_ids = []

    # Use num_experts from topk input shape (authoritative), fall back to max_expert_id
    max_expert_id = max(expert_weights.keys()) if expert_weights else 0
    total_experts = num_experts if num_experts > 0 else (max_expert_id + 1)

    # Derive block/layer identifier from any weight to construct missing expert IDs
    # The weight pattern is e.g., "param::block.1.ffn.expert.0.gate.weight"
    # We need to extract "block.1" to reconstruct missing expert paths
    block_prefix = None
    block_pattern = re.compile(r"param::(.*?)\.expert[s]?\.\d+")
    for projs in expert_weights.values():
        for tid in projs.values():
            m = block_pattern.match(tid)
            if m:
                block_prefix = m.group(1)  # e.g., "block.1.ffn" or "model.layers.1.block_sparse_moe"
                break
        if block_prefix is not None:
            break

    # Get reference shapes from any existing expert (for missing expert tensor entries)
    ref_shapes: Dict[str, list] = {}
    for projs in expert_weights.values():
        for proj_type, tid in projs.items():
            if proj_type not in ref_shapes:
                tdata = tensors.get(tid, {})
                if "shape" in tdata:
                    ref_shapes[proj_type] = tdata["shape"]
        if len(ref_shapes) == 3:
            break

    for expert_id in range(total_experts):
        if expert_id in expert_weights:
            projs = expert_weights[expert_id]
            # Some experts may be partially traced (e.g., gate+up present but down
            # was not activated during trace). Fill missing projections from pattern.
            gate_tid = projs.get("gate") or (
                f"param::{block_prefix}.expert.{expert_id}.gate.weight"
                if block_prefix else "")
            up_tid = projs.get("up") or (
                f"param::{block_prefix}.expert.{expert_id}.up.weight"
                if block_prefix else "")
            down_tid = projs.get("down") or (
                f"param::{block_prefix}.expert.{expert_id}.down.weight"
                if block_prefix else "")
            gate_ids.append(gate_tid)
            up_ids.append(up_tid)
            down_ids.append(down_tid)

            # Ensure tensor entries exist for synthesized IDs
            for proj_type, tid in [("gate", gate_tid), ("up", up_tid), ("down", down_tid)]:
                if tid and tid not in tensors:
                    tensors[tid] = {
                        "shape": ref_shapes.get(proj_type, []),
                        "dtype": "float32",
                        "weight_name": tid[len("param::"):],
                    }
        elif block_prefix is not None:
            # Expert absent from graph (never activated during trace)
            # Construct tensor ID from pattern — weights ARE in the checkpoint
            gate_tid = f"param::{block_prefix}.expert.{expert_id}.gate.weight"
            up_tid = f"param::{block_prefix}.expert.{expert_id}.up.weight"
            down_tid = f"param::{block_prefix}.expert.{expert_id}.down.weight"
            gate_ids.append(gate_tid)
            up_ids.append(up_tid)
            down_ids.append(down_tid)

            # Ensure tensor entries exist so they get arena slots
            for proj_type, missing_tid in [("gate", gate_tid), ("up", up_tid), ("down", down_tid)]:
                if missing_tid not in tensors:
                    tensors[missing_tid] = {
                        "shape": ref_shapes.get(proj_type, []),
                        "dtype": "float32",
                        "weight_name": missing_tid[len("param::"):],  # strip param:: prefix
                    }

    return (
        {"gate": gate_ids, "up": up_ids, "down": down_ids},
        num_experts_found,
        hidden_states_tid,
    )


def _count_total_experts(
    tensors: Dict[str, Any],
    topk_uid: str,
    ops: Dict[str, Any],
) -> int:
    """
    Derive num_experts from the gate_scores tensor shape (dim -1).
    This is the softmax output shape's last dimension.

    Falls back to counting weight tensors if shape unavailable.
    """
    topk_data = ops.get(topk_uid, {})
    input_shapes = topk_data.get("input_shapes", [])
    if input_shapes and len(input_shapes[0]) >= 2:
        # gate_scores shape: [seq_len, num_experts]
        return input_shapes[0][-1]

    # Fallback: count from weight tensor IDs
    weight_pattern = re.compile(r"param::block\.\d+\.ffn\.expert\.(\d+)\.gate\.weight")
    expert_ids = set()
    for tid in tensors:
        m = weight_pattern.match(tid)
        if m:
            expert_ids.add(int(m.group(1)))
    return len(expert_ids) if expert_ids else 0


# ============================================================================
# UTILITY FUNCTIONS
# ============================================================================

def _extract_topk_k(op_data: Dict[str, Any]) -> Optional[int]:
    """Extract k value from topk op attributes."""
    attrs = op_data.get("attributes", {})
    args = attrs.get("args", [])
    # topk args: [input_tensor, k, dim, largest, sorted]
    for arg in args:
        if arg.get("type") == "scalar" and isinstance(arg.get("value"), int):
            return arg["value"]
    # Also check top-level attributes
    return attrs.get("k")


def _get_input_tensor_id(op_data: Dict[str, Any], idx: int) -> Optional[str]:
    """Get tensor_id from op's args at position idx."""
    attrs = op_data.get("attributes", {})
    args = attrs.get("args", [])
    tensor_idx = 0
    for arg in args:
        if arg.get("type") == "tensor":
            if tensor_idx == idx:
                return arg.get("tensor_id")
            tensor_idx += 1
        elif arg.get("type") == "tensor_tuple":
            # For index ops, tensor_tuple contains the index tensors
            if tensor_idx == idx:
                tids = arg.get("tensor_ids", [])
                return tids[0] if tids else None
            tensor_idx += 1
    # Also check input_tensor_ids
    input_tids = op_data.get("input_tensor_ids", [])
    if idx < len(input_tids):
        return input_tids[idx]
    return None


def _build_consumer_map(
    ops: Dict[str, Any],
    execution_order: List[str],
) -> Dict[str, List[str]]:
    """Build map: tensor_id -> list of op_uids that consume it."""
    consumer_map: Dict[str, List[str]] = {}
    for op_uid in execution_order:
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        # Collect all tensor_ids referenced in args
        for tid in _collect_input_tids(op_data):
            if tid not in consumer_map:
                consumer_map[tid] = []
            consumer_map[tid].append(op_uid)
    return consumer_map


def _build_producer_map(
    ops: Dict[str, Any],
    execution_order: List[str],
) -> Dict[str, str]:
    """Build map: tensor_id -> op_uid that produces it."""
    producer_map: Dict[str, str] = {}
    for op_uid in execution_order:
        op_data = ops.get(op_uid)
        if op_data is None:
            continue
        for out_tid in op_data.get("output_tensor_ids", []):
            producer_map[out_tid] = op_uid
    return producer_map


def _collect_input_tids(op_data: Dict[str, Any]) -> List[str]:
    """Collect all input tensor IDs from op's args."""
    tids = []
    attrs = op_data.get("attributes", {})
    args = attrs.get("args", [])
    for arg in args:
        if arg.get("type") == "tensor":
            tid = arg.get("tensor_id")
            if tid:
                tids.append(tid)
        elif arg.get("type") == "tensor_tuple":
            for tid in arg.get("tensor_ids", []):
                tids.append(tid)
    # Also from input_tensor_ids if present
    for tid in op_data.get("input_tensor_ids", []):
        if tid not in tids:
            tids.append(tid)
    return tids



# The code for IndentedBuffer is adapted from
# https://github.com/pytorch/pytorch/blob/ed48ea9997c2b04736096e4b6669543ab2e627d5/torch/_inductor/utils.py#L742
# The code for Namespace is adapted from
# https://github.com/pytorch/pytorch/blob/ed48ea9997c2b04736096e4b6669543ab2e627d5/torch/fx/graph.py#L115

# License from pytorch(https://github.com/pytorch/pytorch)

# From PyTorch:

# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# From Caffe2:

# Copyright (c) 2016-present, Facebook Inc. All rights reserved.

# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.

# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.

# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.

# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain

# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.

# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.

# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.

# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.

# All rights reserved.

import builtins
import contextlib
import keyword
import os
import re
import threading
import uuid
from collections import defaultdict
from io import StringIO
from pathlib import Path
from typing import Dict, Set


class IndentedBuffer:
    tabwidth = 4

    def __init__(self, initial_indent=0):
        self._lines = []
        self._indent = initial_indent

    def getvalue(self) -> str:
        buf = StringIO()
        for line in self._lines:
            assert isinstance(line, str)
            buf.write(line)
            buf.write("\n")
        return buf.getvalue()

    def clear(self):
        self._lines.clear()

    def __bool__(self):
        return bool(self._lines)

    def prefix(self):
        return " " * (self._indent * self.tabwidth)

    def newline(self):
        self.writeline("\n")

    def writeline(self, line):
        if line.strip():
            self._lines.append(f"{self.prefix()}{line}")
        else:
            self._lines.append("")

    def tpl(self, format_str, **kwargs):
        assert isinstance(format_str, str), "format_str must be string of type."
        format_str = format_str.format(**kwargs)
        lines = format_str.strip().splitlines()
        for line in lines:
            line = line.replace("\t", " " * self.tabwidth)
            self.writeline(line)

    def writelines(self, lines):
        for line in lines:
            self.writeline(line)

    def writemultiline(self, s):
        self.writelines(s.splitlines())

    def indent(self, offset=1):
        @contextlib.contextmanager
        def ctx():
            self._indent += offset
            try:
                yield
            finally:
                self._indent -= offset

        return ctx()


class NameSpace:
    def __init__(self):
        self._used_names: Set[str] = set()
        self._base_count: Dict[str, int] = defaultdict(int)

        self._illegal_char_regex = re.compile("[^0-9a-zA-Z_]+")
        self._name_suffix_regex = re.compile(r"(.*)_(\d+)$")

    def create_name(self, candidate: str) -> str:
        """Create a unique name.

        Arguments:
            candidate: used as the basis for the unique name, relevant to the user.
        """
        # delete all characters that are illegal in a Python identifier
        candidate = self._illegal_char_regex.sub("_", candidate)

        if not candidate:
            candidate = "_unnamed"

        if candidate[0].isdigit():
            candidate = f"_{candidate}"

        match = self._name_suffix_regex.match(candidate)
        if match is None:
            base = candidate
            num = None
        else:
            base, num_str = match.group(1, 2)
            num = int(num_str)

        candidate = base if num is None else f"{base}_{num}"
        if not num:
            num = self._base_count[base]

        while candidate in self._used_names or self._is_illegal_name(candidate):
            num += 1
            candidate = f"{base}_{num}"

        self._used_names.add(candidate)
        self._base_count[base] = num
        return candidate

    def _is_illegal_name(self, name: str) -> bool:
        # 1. keywords are never allowed as names.
        if name in keyword.kwlist:
            return True

        # 2. Can't shadow a builtin name, unless you *are* that builtin.
        if name in builtins.__dict__:
            return True

        return False


def write_atomic(
    path_: str,
    content: str,
    make_dirs: bool = False,
    encoding: str = "utf-8",
) -> None:
    path = Path(path_)
    if make_dirs:
        path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = (
        path.parent / f".{os.getpid()}.{threading.get_ident()}.{uuid.uuid4().hex}.tmp"
    )
    with tmp_path.open("wt", encoding=encoding) as f:
        f.write(content)
    tmp_path.replace(path)



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""vLLM copy_blocks."""

from typing import Final

import torch

from conch.kernels.vllm.copy_blocks import copy_blocks_launcher


def _validate_sizes(
    key_caches: list[torch.Tensor],
    value_caches: list[torch.Tensor],
    block_mapping: torch.Tensor,
) -> None:
    """Helper function to validate sizes of input tensors to copy_blocks_launcher.

    Args:
        key_caches: List of key_cache for a set of layers, len: num_layers, each element shape: (num_blocks, cache_block_size * num_kv_heads * head_size).
        value_caches: List of value_cache for a set of layers, len: num_layers, num_layers, each element shape: (num_blocks, cache_block_size * num_kv_heads * head_size).
        block_mapping: List of source/destination pairs, shape: (num_pairs, 2).
    """
    num_layers: Final = len(key_caches)
    if (num_layers_value_cache := len(value_caches)) != num_layers:
        msg = f"Mismatch in number of layers between key_caches ({num_layers}) and value_caches ({num_layers_value_cache})"
        raise ValueError(msg)

    if num_layers == 0:
        msg = "Empty list of kv caches passed to copy_blocks"
        raise ValueError(msg)

    expected_shape: Final = key_caches[0].shape
    expected_kv_cache_dims: Final = 2
    if (actual_len := len(expected_shape)) != expected_kv_cache_dims:
        msg = f"Entry in key cache different-dimensional shape ({actual_len}) than expected ({expected_kv_cache_dims}; shape=(num_cache_blocks, cache_block_size * num_kv_heads * head_size))"
        raise ValueError(msg)

    if any(key_cache.shape != expected_shape for key_cache in key_caches) or any(
        value_cache.shape != expected_shape for value_cache in value_caches
    ):
        msg = "Mismatch in shape of entries in key/value caches"
        raise ValueError(msg)

    expected_stride_0: Final = key_caches[0].stride(0)
    if any(key_cache.stride(0) != expected_stride_0 for key_cache in key_caches) or any(
        value_cache.stride(0) != expected_stride_0 for value_cache in value_caches
    ):
        msg = "Mismatch in stride(0) of entries in key/value caches"
        raise ValueError(msg)

    expected_dtype: Final = key_caches[0].dtype
    if any(key_cache.dtype != expected_dtype for key_cache in key_caches) or any(
        value_cache.dtype != expected_dtype for value_cache in value_caches
    ):
        msg = "Mismatch in dtype of entries in key/value caches"
        raise ValueError(msg)

    expected_block_mapping_dim: Final = 2
    if (block_mapping_dim := len(block_mapping.shape)) != expected_block_mapping_dim:
        msg = f"Block mapping tensor is different-dimensional shape ({block_mapping_dim}) than expected ({expected_block_mapping_dim}; shape=(num_pairs, 2))"
        raise ValueError(msg)

    expected_block_mapping_pair_size: Final = 2
    if block_mapping.size(1) != expected_block_mapping_pair_size:
        msg = f"Block mapping tensor has invalid shape ({block_mapping.shape}), expected shape=(num_pairs, 2))"
        raise ValueError(msg)


def copy_blocks(
    key_caches: list[torch.Tensor],
    value_caches: list[torch.Tensor],
    block_mapping: torch.Tensor,
) -> None:
    """Copy cache blocks from one section of the KV cache to another.

    Args:
        key_caches: List of key_cache for a set of layers, len: num_layers, each element shape: (num_blocks, cache_block_size * num_kv_heads * head_size).
        value_caches: List of value_cache for a set of layers, len: num_layers, num_layers, each element shape: (num_blocks, cache_block_size * num_kv_heads * head_size).
        block_mapping: Tensor in form of list of source/destination pairs, shape: (num_pairs, 2).
    """
    # Verify input sizes/tensor shapes
    _validate_sizes(key_caches, value_caches, block_mapping)

    # Call kernel launch wrapper
    copy_blocks_launcher(key_caches, value_caches, block_mapping)



"""
NeuroBrix Validators - Fusion Validator.

Validates kernel fusion correctness.

ZERO FALLBACK: Invalid fusions raise explicit FusionError.
"""

from typing import Any, Dict, List, Set, Tuple, Optional

from .base import BaseValidator, SafetyLevel, ValidationResult, FusionError
from .config import get_config


class FusionValidator(BaseValidator):
    """
    Validates kernel fusion correctness.

    Checks:
    - Fusion candidate validity
    - Data dependency preservation
    - Memory access patterns
    - Fusion size limits
    - Shape compatibility
    """

    # Operations that can be fused together
    FUSIBLE_OPS: Set[str] = {
        # Elementwise ops (highly fusible)
        'add', 'sub', 'mul', 'div', 'sqrt', 'exp', 'log', 'pow',
        'relu', 'gelu', 'silu', 'sigmoid', 'tanh',
        # Reduction ops (limited fusion)
        'sum', 'mean', 'max', 'min',
        # Normalization (can fuse with elementwise)
        'layernorm', 'rmsnorm',
    }

    # Operations that cannot be fused
    UNFUSIBLE_OPS: Set[str] = {
        'matmul', 'linear', 'conv2d', 'conv1d',
        'scaled_dot_product_attention', 'flash_attention',
    }

    # Operations that break fusion chains
    FUSION_BREAKERS: Set[str] = {
        'concat', 'split', 'reshape', 'transpose', 'permute',
        'gather', 'scatter',
    }

    @property
    def name(self) -> str:
        return "FusionValidator"

    def __init__(self, safety_level: SafetyLevel = SafetyLevel.STANDARD):
        super().__init__(safety_level)
        self.config = get_config()

    def validate(self, graph: Any, context: Dict[str, Any]) -> ValidationResult:
        """
        Validate fusion plan for the graph.

        Args:
            graph: Execution graph with nodes
            context: Must contain:
                - 'fusion_groups': List of node groups to fuse

        Returns:
            ValidationResult with fusion validation status
        """
        warnings = []
        errors = []
        metrics = {}

        fusion_groups = context.get('fusion_groups', [])

        if not fusion_groups:
            return ValidationResult(
                passed=True,
                validator_name=self.name,
                message="No fusion groups to validate",
                warnings=warnings,
            )

        metrics['num_fusion_groups'] = len(fusion_groups)

        total_ops = 0
        for i, group in enumerate(fusion_groups):
            group_errors = self._validate_fusion_group(group, graph, context)
            for error in group_errors:
                errors.append(f"Fusion group {i}: {error}")
            total_ops += len(group)

        metrics['total_fused_ops'] = total_ops
        metrics['avg_group_size'] = total_ops / len(fusion_groups) if fusion_groups else 0

        if errors:
            return ValidationResult(
                passed=False,
                validator_name=self.name,
                message="\n".join(errors),
                warnings=warnings,
                metrics=metrics,
            )

        return ValidationResult(
            passed=True,
            validator_name=self.name,
            message=f"Validated {len(fusion_groups)} fusion groups with {total_ops} ops",
            warnings=warnings,
            metrics=metrics,
        )

    def _validate_fusion_group(
        self,
        group: List[Any],
        graph: Any,
        context: Dict[str, Any],
    ) -> List[str]:
        """Validate a single fusion group."""
        errors = []
        config = self.config

        # Check group size
        if len(group) > config.max_fusion_ops:
            errors.append(
                f"Group size {len(group)} exceeds max {config.max_fusion_ops}"
            )

        # Check ops are fusible
        group_ops = []
        for node in group:
            op = getattr(node, 'op', None)
            if op is None:
                continue
            op_lower = op.lower()
            group_ops.append(op_lower)

            if op_lower in self.UNFUSIBLE_OPS:
                errors.append(f"Op '{op}' cannot be fused")

        # Check for fusion breakers in the middle
        if len(group) > 2:
            middle_ops = group_ops[1:-1]
            for op in middle_ops:
                if op in self.FUSION_BREAKERS:
                    errors.append(f"Fusion breaker '{op}' in middle of group")

        # Check data dependencies
        dep_errors = self._check_dependencies(group)
        errors.extend(dep_errors)

        # Check shape compatibility
        shape_errors = self._check_shape_compatibility(group, context)
        errors.extend(shape_errors)

        return errors

    def _check_dependencies(self, group: List[Any]) -> List[str]:
        """Check that data dependencies are preserved in fusion."""
        errors = []

        # Build set of nodes in group
        group_names = set()
        for node in group:
            name = getattr(node, 'name', str(node))
            group_names.add(name)

        # Check each node's inputs
        for node in group:
            inputs = getattr(node, 'inputs', [])
            for inp in inputs:
                inp_name = getattr(inp, 'name', str(inp))

                # If input is from outside group, check it comes first
                if inp_name not in group_names:
                    # External dependency - OK, will be handled by fusion kernel
                    pass
                else:
                    # Internal dependency - must be earlier in group
                    found_producer = False
                    for i, other in enumerate(group):
                        other_name = getattr(other, 'name', str(other))
                        if other_name == inp_name:
                            found_producer = True
                            break
                        if other_name == getattr(node, 'name', ''):
                            # Consumer before producer
                            errors.append(
                                f"Dependency violation: {inp_name} must come before "
                                f"{getattr(node, 'name', node)}"
                            )
                            break

        return errors

    def _check_shape_compatibility(
        self,
        group: List[Any],
        context: Dict[str, Any],
    ) -> List[str]:
        """Check that shapes are compatible for fusion."""
        errors = []

        input_shapes = context.get('input_shapes', {})
        if not input_shapes:
            return errors

        # Get shapes of all nodes in group
        shapes = []
        for node in group:
            name = getattr(node, 'name', str(node))
            if name in input_shapes:
                shapes.append(input_shapes[name])

        if len(shapes) < 2:
            return errors

        # Check all shapes are broadcastable
        ref_shape = shapes[0]
        for shape in shapes[1:]:
            if not self._are_broadcastable(ref_shape, shape):
                errors.append(
                    f"Shapes not compatible for fusion: {ref_shape} vs {shape}"
                )

        return errors

    def _are_broadcastable(
        self,
        shape1: Tuple[int, ...],
        shape2: Tuple[int, ...],
    ) -> bool:
        """Check if two shapes are broadcastable."""
        # Align from right
        len1, len2 = len(shape1), len(shape2)
        max_len = max(len1, len2)

        for i in range(1, max_len + 1):
            d1 = shape1[-i] if i <= len1 else 1
            d2 = shape2[-i] if i <= len2 else 1

            if d1 != d2 and d1 != 1 and d2 != 1:
                return False

        return True

    def can_fuse(
        self,
        op1: str,
        op2: str,
    ) -> Tuple[bool, str]:
        """
        Check if two operations can be fused.

        Returns:
            (can_fuse, reason)
        """
        op1_lower = op1.lower()
        op2_lower = op2.lower()

        # Unfusible ops
        if op1_lower in self.UNFUSIBLE_OPS:
            return False, f"'{op1}' cannot be fused"
        if op2_lower in self.UNFUSIBLE_OPS:
            return False, f"'{op2}' cannot be fused"

        # Fusion breakers
        if op1_lower in self.FUSION_BREAKERS:
            return False, f"'{op1}' breaks fusion chains"
        if op2_lower in self.FUSION_BREAKERS:
            return False, f"'{op2}' breaks fusion chains"

        # Both fusible
        if op1_lower in self.FUSIBLE_OPS and op2_lower in self.FUSIBLE_OPS:
            return True, "Both ops are fusible"

        return False, f"Unknown fusion compatibility: {op1}, {op2}"

    def find_fusion_groups(
        self,
        graph: Any,
    ) -> List[List[Any]]:
        """
        Find valid fusion groups in the graph.

        Returns list of node groups that can be fused.
        """
        groups = []
        current_group: List[Any] = []

        nodes = getattr(graph, 'nodes', [])
        if hasattr(graph, '__iter__') and not nodes:
            nodes = list(graph)

        for node in nodes:
            op = getattr(node, 'op', None)
            if op is None:
                continue

            op_lower = op.lower()

            # Check if can add to current group
            if op_lower in self.FUSIBLE_OPS:
                if len(current_group) < self.config.max_fusion_ops:
                    current_group.append(node)
                else:
                    # Group full, start new one
                    if len(current_group) > 1:
                        groups.append(current_group)
                    current_group = [node]
            else:
                # Non-fusible op, close current group
                if len(current_group) > 1:
                    groups.append(current_group)
                current_group = []

        # Don't forget last group
        if len(current_group) > 1:
            groups.append(current_group)

        return groups

    def validate_fusion(
        self,
        nodes: List[Any],
    ) -> None:
        """
        Validate that a list of nodes can be fused.

        Raises FusionError if fusion is invalid.
        """
        if len(nodes) > self.config.max_fusion_ops:
            raise FusionError(
                f"Fusion group too large: {len(nodes)} > {self.config.max_fusion_ops}",
                nodes_to_fuse=[getattr(n, 'name', str(n)) for n in nodes],
                reason="Exceeds max fusion ops limit",
            )

        for node in nodes:
            op = getattr(node, 'op', None)
            if op and op.lower() in self.UNFUSIBLE_OPS:
                raise FusionError(
                    f"Cannot fuse unfusible op: {op}",
                    nodes_to_fuse=[getattr(n, 'name', str(n)) for n in nodes],
                    reason=f"Op '{op}' cannot be fused",
                )

    def should_run(self, safety_level: SafetyLevel) -> bool:
        """Fusion validation runs at STANDARD and PARANOID levels."""
        return safety_level >= SafetyLevel.STANDARD



# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""GeLU Tanh and Mul."""

import torch

from conch.kernels.activation.gelu_tanh_and_mul import gelu_tanh_and_mul_launcher


def gelu_tanh_and_mul(x: torch.Tensor) -> torch.Tensor:
    """gelu_tanh_and_mul operation."""
    d = x.shape[-1] // 2
    output_shape = x.shape[:-1] + (d,)
    output = torch.empty(output_shape, dtype=x.dtype, device=x.device)

    gelu_tanh_and_mul_launcher(output, x)

    return output



import logging

import torch
import triton
import triton.language as tl

from flag_gems.runtime import device, torch_device_fn
from flag_gems.utils import libentry

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
device_ = device


@libentry()
@triton.jit
def eye_kernel(
    out_ptr,
    N,
    M,
    BLOCK_i: tl.constexpr,
    BLOCK_j: tl.constexpr,
):
    pid_i = tl.program_id(0)  # block id
    off_i = pid_i * BLOCK_i + tl.arange(0, BLOCK_i)
    mask_i = off_i < N

    pid_j = tl.program_id(1)  # block id
    off_j = pid_j * BLOCK_j + tl.arange(0, BLOCK_j)
    mask_j = off_j < M

    val = tl.where(off_i[:, None] == off_j[None, :], 1.0, 0.0)
    mask = mask_i[:, None] & mask_j[None, :]
    off_ij = off_i[:, None] * M + off_j[None, :]

    tl.store(out_ptr + off_ij, val, mask=mask)


def eye_m(n, m, *, dtype=None, layout=torch.strided, device=None, pin_memory=None):
    """
    Triton-based implementation of torch.eye_m(n, m), using 2D tiles to split the matrix into blocks.
    """
    logger.debug("GEMS EYE_M")
    if dtype is None:
        dtype = torch.get_default_dtype()
    if device is None:
        device = torch.device(device_.name)
    if layout != torch.strided:
        raise ValueError("Currently only strided layout is supported for eye_m.")

    out = torch.empty(
        (n, m), dtype=dtype, device=device, layout=layout, pin_memory=pin_memory
    )
    BLOCK_SIZE = 32
    grid = (triton.cdiv(n, BLOCK_SIZE), triton.cdiv(m, BLOCK_SIZE))

    with torch_device_fn.device(device):
        eye_kernel[grid](
            out,
            n,
            m,
            BLOCK_SIZE,
            BLOCK_SIZE,
        )
    return out



"""
NeuroBrix Validators - Configuration.

Global configuration for validation behavior.
Thread-safe singleton pattern for consistent settings.

ZERO HARDCODING: All thresholds are configurable.
"""

from dataclasses import dataclass, field
from typing import Dict, Optional, Set
import threading

from .base import SafetyLevel


@dataclass
class ValidatorConfig:
    """
    Configuration for all validators.

    Configurable thresholds for:
    - Memory limits
    - Numerical stability
    - Shape constraints
    - Performance budgets
    """

    # Safety level
    safety_level: SafetyLevel = SafetyLevel.STANDARD

    # Memory thresholds
    memory_safety_margin: float = 0.1  # 10% safety margin
    max_tensor_bytes: int = 16 * 1024**3  # 16 GB max single tensor
    warn_memory_usage: float = 0.8  # Warn at 80% usage

    # Numerical thresholds
    nan_threshold: int = 0  # Any NaN is an error
    inf_threshold: int = 0  # Any Inf is an error
    max_fp16_value: float = 65504.0  # FP16 max representable
    max_bf16_value: float = 3.38953e38  # BF16 max representable
    gradient_clip_threshold: float = 1.0  # Gradient clipping threshold

    # Shape constraints
    max_batch_size: int = 4096
    max_sequence_length: int = 1048576  # 1M tokens
    max_hidden_dim: int = 65536
    max_tensor_dims: int = 8  # Max number of dimensions

    # Sync thresholds
    max_stream_depth: int = 32  # Max concurrent streams
    sync_timeout_ms: int = 30000  # 30 second timeout

    # Fusion thresholds
    max_fusion_ops: int = 16  # Max ops in single fused kernel
    min_fusion_benefit: float = 0.1  # 10% speedup required

    # Supported operations (can be extended)
    supported_ops: Set[str] = field(default_factory=lambda: {
        # Core arithmetic
        "add", "sub", "mul", "div", "matmul", "linear",
        # Activations
        "relu", "gelu", "silu", "sigmoid", "tanh", "softmax",
        # Normalization
        "layernorm", "rmsnorm", "groupnorm", "batchnorm",
        # Reduction
        "sum", "mean", "max", "min",
        # Attention
        "scaled_dot_product_attention", "flash_attention",
        # Convolution
        "conv2d", "conv1d",
        # Pooling
        "maxpool2d", "avgpool2d",
        # Memory
        "reshape", "transpose", "permute", "view", "contiguous",
        "concat", "split", "slice", "gather", "scatter",
        # Embedding
        "embedding", "rotary_embedding",
        # Misc
        "dropout", "sqrt", "exp", "log", "pow",
        # ONNX CamelCase variants
        "scaleddotproductattention", "flashattention", "layernormalization",
    })

    # Performance budgets (optional)
    max_kernel_time_ms: Optional[float] = None
    max_memory_bandwidth_usage: Optional[float] = None

    def validate_memory_requirement(self, required_bytes: int, available_bytes: int) -> bool:
        """Check if memory requirement is within limits."""
        usable = available_bytes * (1.0 - self.memory_safety_margin)
        return required_bytes <= usable

    def is_op_supported(self, op: str) -> bool:
        """Check if operation is supported."""
        return op.lower() in self.supported_ops

    def get_dtype_max(self, dtype: str) -> float:
        """Get max representable value for dtype."""
        dtype_maxes = {
            "float16": self.max_fp16_value,
            "bfloat16": self.max_bf16_value,
            "float32": 3.4028235e38,
            "float64": 1.7976931348623157e308,
        }
        return dtype_maxes.get(dtype.lower(), float('inf'))


# ============================================================================
# Global Configuration Singleton
# ============================================================================

_config_lock = threading.Lock()
_global_config: Optional[ValidatorConfig] = None


def get_config() -> ValidatorConfig:
    """
    Get the global validator configuration.

    Thread-safe singleton pattern.
    """
    global _global_config

    if _global_config is None:
        with _config_lock:
            if _global_config is None:
                _global_config = ValidatorConfig()

    return _global_config


def set_config(config: ValidatorConfig) -> None:
    """
    Set the global validator configuration.

    Thread-safe.
    """
    global _global_config

    with _config_lock:
        _global_config = config


def reset_config() -> None:
    """Reset to default configuration."""
    global _global_config

    with _config_lock:
        _global_config = ValidatorConfig()


def update_config(**kwargs) -> ValidatorConfig:
    """
    Update specific configuration values.

    Returns the updated config.
    """
    config = get_config()

    for key, value in kwargs.items():
        if hasattr(config, key):
            setattr(config, key, value)
        else:
            raise ValueError(f"Unknown config key: {key}")

    return config



import torch, math

N_inp = 64
N_out = 64
d = 128

Q = torch.randn(N_out, d)
K = torch.randn(N_inp, d)
V = torch.randn(N_inp, d)

O = torch.zeros(N_out, d)  # Output matrix initialized to zeros
L = torch.zeros(N_out, 1)  # Logsumexp values initialized to zeros

B_c = 16  # Column block size
B_r = 16  # Row block size

T_c = (N_inp + B_c - 1) // B_c  # Number of column tiles
T_r = (N_out + B_r - 1) // B_r  # Number of row tiles

scale_factor = 1 / math.sqrt(d)  # Scaling factor for softmax stability

for i in range(T_r):
    Q_i = Q[i * B_r : (i + 1) * B_r]
    O_i = torch.zeros(B_r, d)
    l_i = torch.zeros(B_r, 1)
    m_i = torch.full((B_r, 1), -math.inf)  # Initialize max value to a very low number

    # Iterate over tiles of K and V
    for j in range(T_c):
        K_j = K[j * B_c : (j + 1) * B_c]
        V_j = V[j * B_c : (j + 1) * B_c]

        # Compute the scaled attention scores (S_i = Q_i @ K_j.T)
        S_i = scale_factor * (Q_i @ K_j.T)

        # Update m_i for numerical stability (max of previous and current blocks)
        new_m_i = torch.maximum(m_i, S_i.max(dim=-1, keepdim=True).values)

        # Compute P_i using the exponent of the stabilized scores
        P_i = torch.exp(S_i - new_m_i)

        # Update l_i (softmax denominator) with the new block values
        l_i = torch.exp(m_i - new_m_i) * l_i + P_i.sum(dim=-1, keepdim=True)

        # Update O_i (the output matrix) using the attention weights
        O_i = torch.exp(m_i - new_m_i) * O_i + P_i @ V_j

        # Update m_i after processing the block
        m_i = new_m_i

    O_i = O_i / l_i  # Normalize O_i by l_i for the softmax output

    # Update output and logsumexp values for this block
    O[i * B_r : (i + 1) * B_r] = O_i
    L[i * B_r : (i + 1) * B_r] = m_i + torch.log(
        l_i
    )  # Store the logsumexp for backprop

if __name__ == "__main__":
    expected = torch.nn.functional.scaled_dot_product_attention(Q, K, V)
    print("Max absolute difference: ", (O - expected).abs().max())



# REVISED FROM
# https://github.com/shawntan/stickbreaking-attention/blob/main/stickbreaking_attention/sb_varlen/softplus.py

import triton
from triton import language as tl

from fla.utils import IS_NVIDIA


def _generate_softplus(num_pack):
    template = """
        .reg .pred p;
        setp.gt.f32  p, ${in_reg}, 20.;
        @p  mov.f32  ${out_reg}, ${in_reg};
        @!p mul.f32            ${out_reg}, ${in_reg}, 1.4426950408889634;
        @!p ex2.approx.ftz.f32 ${out_reg}, ${out_reg};
        @!p add.f32            ${out_reg}, ${out_reg}, 1.0;
        @!p lg2.approx.ftz.f32 ${out_reg}, ${out_reg};
        @!p mul.f32            ${out_reg}, ${out_reg}, 0.6931471805599453;
    """
    out_str = ""

    for i in range(num_pack):
        inner_str = template.format(out_reg=i, in_reg=i + num_pack)
        out_str += "{" + inner_str + "}\n"
    # flatten out because torch.compile doesn't like newlines
    out_str = " ".join(out_str.split("\n"))
    return out_str


def _generate_softplus2(num_pack):
    template = """
        .reg .pred p;
        setp.gt.f32  p, ${in_reg}, 15.;
        @p  mov.f32  ${out_reg}, ${in_reg};
        @!p ex2.approx.ftz.f32 ${out_reg}, ${in_reg};
        @!p add.f32            ${out_reg}, ${out_reg}, 1.0;
        @!p lg2.approx.ftz.f32 ${out_reg}, ${out_reg};
    """
    out_str = ""

    for i in range(num_pack):
        inner_str = template.format(out_reg=i, in_reg=i + num_pack)
        out_str += "{" + inner_str + "}\n"
    # flatten out because torch.compile doesn't like newlines
    out_str = " ".join(out_str.split("\n"))
    return out_str


def _generate_constraints(num_pack):
    return ",".join("=r" for i in range(num_pack)) + "," + ",".join("r" for i in range(num_pack))


_NUM_REG = 1
s_softplus: tl.constexpr = tl.constexpr(_generate_softplus(_NUM_REG))
s_softplus2: tl.constexpr = tl.constexpr(_generate_softplus2(_NUM_REG))
s_constraints: tl.constexpr = tl.constexpr(_generate_constraints(_NUM_REG))
NUM_REG: tl.constexpr = tl.constexpr(_NUM_REG)


@triton.jit
def softplus_nv(x):
    # equivalent to:
    # return tl.where(x < 20.0, tl.math.log(1 + tl.math.exp(x)), x)
    return tl.inline_asm_elementwise(
        asm=s_softplus,
        constraints=s_constraints,
        pack=NUM_REG,
        args=[
            x,
        ],
        dtype=tl.float32,
        is_pure=True,
    )

@triton.jit
def softplus_triton(x):
    return tl.where(x < 20.0, tl.math.log(1 + tl.math.exp(x)), x)

@triton.jit
def softplus2_nv(x):
    # equivalent to:
    # return tl.where(x < 15.0, tl.math.log2(1 + tl.math.exp2(x)), x)
    return tl.inline_asm_elementwise(
        asm=s_softplus2,
        constraints=s_constraints,
        pack=NUM_REG,
        args=[
            x,
        ],
        dtype=tl.float32,
        is_pure=True,
    )

@triton.jit
def softplus2_triton(x):
    return tl.where(x < 15.0, tl.math.log2(1 + tl.math.exp2(x)), x)

if IS_NVIDIA:
    softplus = softplus_nv
    softplus2 = softplus2_nv
else:
    softplus = softplus_triton
    softplus2 = softplus2_triton
